#include "../config.h"
#include "../src/drip.h"
extern "C" {
#include "../libvo/yuv2rgb.h"
}

/* Lookup tables */
gfloat T1164[256];
gfloat T2018[256];
gfloat T0813[256];
gfloat T1596[256];
gfloat T0391[256];
/* Init done flag */
static gboolean init_done = FALSE;


/* Limit colour to valid range */
inline guint8 normalize_colour(gfloat colour) {
    static guint colour_norm;

    if (colour<0) {
        colour_norm = 0;
    } else if (colour>255) {
        colour_norm = 255;
    } else {
        colour_norm = (guint8)(colour);
    }
    return colour_norm;
}
       

/* TODO no chroma interpolating is done atm */
void I420toRGB24(guint8 *dst, guint8 *src, gint width, gint height) {
    register gint offset = 0;
    register gint offset4 = 0;
    register gint offset3 = 0;
    static gfloat offset4_f;
    register guint8 Y,CR,CB;
    register gfloat Bf,Gf,Rf;
    static guint8 *src0,*src1,*src2;
 
    src0 = src;
    src1 = (guint8*)((glong)src0+width*height);
    src2 = (guint8*)((glong)src1+width*height/4);

    offset = 0;
    offset3 = 0;
    /* Loop */
    for (register gint y=0;y<height;y++) {
        offset4_f = (y>>1) * (width>>1);
        for (register gint x=0;x<width;x++) {
             offset4 = (gint)offset4_f;

             Y  = src0[offset];
             CB = src1[offset4];
             CR = src2[offset4];
 
             Bf = T1164[Y] + T2018[CB];
             Gf = T1164[Y] - T0813[CR] - T0391[CB];
             Rf = T1164[Y] + T1596[CR];
 
             dst[offset3++] = normalize_colour(Rf); //(B) + (G<<8) + (R<<16); // RGB 24bits
             dst[offset3++] = normalize_colour(Gf);
             dst[offset3++] = normalize_colour(Bf);
 
             offset++;
             offset4_f = offset4_f+0.5;
        }
    }

    /* Clean & Exit */
    return;
}   

void I420toRGB24_init(void) {
    /* Calculate lookup tables */
    for (gfloat i=-16;i<240;i++) {
        T1164[(gint)(i+16)] = i*1.164;
    }
    for (gfloat i=-128;i<128;i++) {
        T2018[(gint)(i+128)] = i*2.018;
        T0813[(gint)(i+128)] = i*0.813;
        T1596[(gint)(i+128)] = i*1.596;
        T0391[(gint)(i+128)] = i*0.391;
    }
    init_done = TRUE;
    return;
}

