/* Mpeg3 decoding */
#include "demuxer.hh"
#include "encoder.hh"
/* SPU */
#include "../libspu/spu.h"
/* Other */
#include "plugin-loader.hh"
#include "stablelizer.hh"
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "fast_memcpy.hh"


/* Defines */
#define DEMUX_BUFFER_SIZE  16*65536
#define AUDIO_BUFFER_SIZE 128*65536
#define PTSMAX 8192
//#define DEBUG

/* Globals */
static mpeg2dec_t *mpeg2dec;
static vo_open_t * video_output_open = NULL;
static ao_open_t * audio_output_open = NULL;
static vo_instance_t *video_output;
static vo_open_t *video_driver = NULL;
static ao_instance_t *audio_output;
static uint8_t buffer[AUDIO_BUFFER_SIZE];
static FILE *in_file;
static gint width = -1,height=-1;
static gfloat framerate;
static sample_t *output_liba52;
static glong total_frames = 0;
static gboolean demuxing = FALSE;
static gboolean demuxing_frames = FALSE;
static gboolean process = FALSE;
static void(*processor)(guint8*[3]);
static gint16 gint16_samples[6*256*6*sizeof(gint16)];
static gint16 *audio_buffer = NULL;
static gint audio_buffer_index = 0;
static gint audio_buffer_base = 0;
static gint audio_buffer_total = 0;
static gint audio_fetched_total = 0;
static gint audio_sample_rate,audio_bit_rate,audio_channels;
static int demux_pid = 0;
glong SCRstream, SCRvideo, PTSprivate;
glong SCRstream_prev = 0;
static glong PTStable[PTSMAX];
static glong BLOCKSIZE[PTSMAX];
static glong IDtable[PTSMAX];
static glong IDnumber = 0;
static gint PTSindex = 0;
static gint PTSvideo = -1;
gulong PTSframe; // Here PTS value of latest video frame is stored.. or should be, not 100% garanteed for now, couse libmpeg2 cant supply the correct value
static gint16 silence[512];
static gint16 audio_buffer_chunk[2048];
static gint16 audio_buffer_stable[32768];
static gfloat AUDIO_BYTES_PER_FRAME = -1;//drip_audio_samplerate() / drip_mpeg_framerate() * 4;
static gfloat PTS_PER_AUDIO_FRAME = 480;
static gfloat PTS_PER_AUDIO_BYTE = PTS_PER_AUDIO_FRAME/1024;
common_instance_t *vo_open_instance;
gchar **input_files;
gint input_files_index = 0;
gint spu_track;
gint audio_track;
glong avifile_audio_outstream_bytes = 0;



/* ----------- Functions -- */

/* Return audio samplerate */
gint drip_audio_samplerate(void) {
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"drip_audio_samplerate = %i\n",audio_sample_rate);
    switch (audio_sample_rate) {
        case 0:  return 48000;
                 break;
        default: return audio_sample_rate;
                 break;
    }
}


/* Return audio bitrate */
gint drip_audio_bitrate(void) {
    return audio_bit_rate;
}


/* Return audio bitrate */
gint drip_audio_channels(void) {
    return audio_channels;
}


/* Return width of current open mpeg */
gint drip_mpeg_width(void) {
    return width;
}


/* Return height of current open mpeg */
gint drip_mpeg_height(void) {
    return height;
}


/* Determine framerate out frame_rate_code */
inline void drip_framerate(mpeg2dec_t* _mpeg2dec) {
    switch (_mpeg2dec->frame_rate_code) {
    case 1: framerate = 23.976;
            break;
    case 2: framerate = 24;
            break;
    case 3: framerate = 25;
            break;
    case 4: framerate = 29.97;
            break;
    case 5: framerate = 30;
            break;
    case 6: framerate = 50;
            break;
    case 7: framerate = 59.94;
            break;
    case 8: framerate = 60;
            break;
    default:framerate = -1;
            break;
    }
}


/* Return framerate of current open mpeg */
gfloat drip_mpeg_framerate(void) {
    return framerate;
}


/* Return TimeStamp value per video frame */
gfloat TimeStamp_Frame(void) {
    return 90000/drip_mpeg_framerate();
}


/* Seek to percentage in mpeg file */
gint drip_mpeg_seek_percentage(gfloat perc) {
    gint result,pos;
    /* Check perc */
    if (perc<0 || perc>1) {
        perc = 0;
    }
    /* Calc position in bytes from perc */
    result = fseek(in_file,0,SEEK_END);
    pos = (gint)(ftell(in_file) * perc);
    /* Seek to pos */
    result = fseek(in_file,pos,SEEK_SET);
    return result;
}


/* float_to_int */
static inline gint16 convert(int32_t i) {
    if (i>0x43c07fff)
        return 32767;
    else if (i<0x43bf8000)
        return -32768;
    else
        return i-0x43c00000;
}


/* Typecast floats to int16s */
static inline gint float_to_int_stereo(gfloat *_f, gint16 *s16) {
    gint i;
    int32_t *f = (int32_t*)_f;       /* XXX assumes IEEE float format */
    for (i=0; i<256; i++) {
        s16[i*2] = convert(f[i]);
        s16[i*2+1] = convert(f[i+256]);
    }
    return 256*2*(sizeof(gint16));
}


/* Reset audio buffer */
inline void reset_audio_buffer() {
    audio_buffer_index = 0;
    PTSindex = 0;
    //PTSbase = 0;
}


/* Debug audio buffer, display status */
inline void audio_buffer_status(void) {
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"index=%i    base=%i   fetched=%i    stored=%i    bufferd=%i\n",audio_buffer_index,audio_buffer_base,audio_fetched_total,audio_buffer_total,audio_buffer_total-audio_fetched_total);
}


/* Returns audio buffer and resets index */
gint16* fetch_from_audio_buffer(gint* samples, gint SCR) {
    gint chunk_size = BLOCKSIZE[0]; // TODO: 2->channels
    gint i;
    guint min;

    /* Set samples for calling function */
    *samples = BLOCKSIZE[0];

    /* Anything in buffer? */
    if ((audio_buffer_index==0) || (PTSindex<1)) {
        #ifdef DEBUG 
        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Audio buffer empty (%i audio frames)",chunk_size,PTSindex);
        #endif
        return NULL;
    }

    /* PTS value passed? */
    if ((SCR<PTStable[0] && ((glong)PTStable[0]-(glong)SCR)<50000)) {
        #ifdef DEBUG
        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Audio buffer holds data that doesn't need to be played yet (requested PTS=%i, audio PTS=%i)",SCR,PTStable[0]);
        #endif
        *samples = 512;
        return silence;
    }

    /* Enough in buffer? */
    if (audio_buffer_index<chunk_size) {
        #ifdef DEBUG
        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Audio buffer holds %ibytes, could not release %ibytes",audio_buffer_index,chunk_size);
        #endif
        return NULL;
    }
    /* Moronic chuck_size? */
    if (chunk_size<1) {
        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Audio buffer request for %ib canceled",chunk_size);
        return NULL;
    }

    /* Keep track total bytes put here */
    audio_fetched_total+=chunk_size;

    /* Update buffer */
    // TODO add mutex locking here and in add_to_audio_buffer(..)
    audio_buffer_index -= chunk_size;
    if(audio_buffer_index<0) {
        audio_fetched_total += audio_buffer_index;
        audio_buffer_index=0;
        #ifdef DEBUG
        g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"Audio buffer underflow");
        #endif
    }
    #ifdef DEBUG
    audio_buffer_status();
    #endif

    #ifdef DEBUG
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Audio buffer returned audio sample frame (requested PTS=%i, audio PTS=%i, ID=%i)",SCR,PTStable[0],IDtable[0]);
    #endif

    /* Update PTStable, BLOCKSIZE table and PTSindex */
    fast_memcpy(PTStable,(gint16*)((glong)PTStable+sizeof(glong)),sizeof(glong)*(PTSMAX-1));
    fast_memcpy(BLOCKSIZE,(gint16*)((glong)BLOCKSIZE+sizeof(glong)),sizeof(glong)*(PTSMAX-1));
    fast_memcpy(IDtable,(gint16*)((glong)IDtable+sizeof(glong)),sizeof(glong)*(PTSMAX-1));
    PTSindex--;
    /* Copy audio data into return pointer */
    fast_memcpy(audio_buffer_chunk, (gint16*)((glong)audio_buffer+(glong)audio_buffer_base),chunk_size);
    /* Update buffer base (offset) */
    audio_buffer_base += chunk_size;

    /* Stablelizer filter */
    *samples = stablelizer_audio((guint8*)audio_buffer_chunk,(guint8*)audio_buffer_stable,(gulong)*samples,(gulong)SCR);
    return audio_buffer_stable;
}


/* Builds audio buffer for encoder */
void add_to_audio_buffer(gint16* audio_chunk, gint audio_chunk_size, gint PTS) {
    static gint16* audio_buffer_dest;
    static gint i;
    static gint PTSoff = 0;

    /* Process audio data by the filters */
    /* - Plugin Filters, type AUDIO PRE - */
    for (gint i=0;i<filters_index;i++) {
        if (filters[i].active==TRUE & filters[i].module_type==AUDIO && filters[i].module_phase==PRE) {
            audio_chunk = (gint16*)(*filters[i].module_apply_function)((guint8*)audio_chunk,audio_chunk_size,(gulong)SCRdrip,PTS);
        }
    }

    /* - Plugin Filters, type AUDIO NORMAL - */
    for (gint i=0;i<filters_index;i++) {
        if (filters[i].active==TRUE & filters[i].module_type==AUDIO && filters[i].module_phase==NORMAL) {
            audio_chunk = (gint16*)(*filters[i].module_apply_function)((guint8*)audio_chunk,audio_chunk_size,(gulong)SCRdrip,PTS);
        }
    }

    /* - Plugin Filters, type AUDIO POST - */
    for (gint i=0;i<filters_index;i++) {
        if (filters[i].active==TRUE & filters[i].module_type==AUDIO && filters[i].module_phase==POST) {
            audio_chunk = (gint16*)(*filters[i].module_apply_function)((guint8*)audio_chunk,audio_chunk_size,(gulong)SCRdrip,PTS);
        }
    }


    /* Set PTSoff to PTS */
    if (PTSoff==0) PTSoff = PTS;

    /* Keep track total bytes put here */
    audio_buffer_total+=audio_chunk_size;
    /* 1st time here? */
    if (audio_buffer==NULL) {
        audio_buffer = (gint16*)malloc(AUDIO_BUFFER_SIZE);
    }
    /* Space left in buffer? */
    if (((glong)audio_buffer_index+(glong)audio_chunk_size) > (AUDIO_BUFFER_SIZE-1)) {
        #ifdef DEBUG
        g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"Audio buffer overflow, resetting audio buffer");
        #endif
        reset_audio_buffer();
        return;
    }

    /* Need to rotate buffer? */
    if (((glong)audio_buffer_index+(glong)audio_chunk_size+(glong)audio_buffer_base) > (AUDIO_BUFFER_SIZE-1)) {
        audio_buffer_base = 0;
        fast_memcpy(audio_buffer,(gint16*)((glong)audio_buffer+(glong)audio_buffer_base),(glong)audio_buffer_index);
        #ifdef DEBUG
        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Audio buffer rotated base");
        #endif
    }

    /* Copy audio chunk into buffer and update index */
    audio_buffer_dest = (gint16*)((glong)audio_buffer + (glong)audio_buffer_base + (glong)audio_buffer_index);
    fast_memcpy(audio_buffer_dest,audio_chunk,(size_t)audio_chunk_size);
    audio_buffer_index += audio_chunk_size;

    /* Update PTS and blcoksize table */
    PTStable[PTSindex] = PTS;
    BLOCKSIZE[PTSindex] = audio_chunk_size;
    IDtable[PTSindex] = IDnumber; 
    IDnumber++;
    PTSindex++;

    #ifdef DEBUG
    audio_buffer_status();
    #endif
    return;
}


/* Add audio data to output */
void output_audio(glong frame) {
    static glong ABO = 0;
    static glong frame_prev = -1;
    gint frame_delta;
    gfloat _frame;
    static gfloat SCR = -1;
    glong SCR33b;
    gint16* audio_chunk;
    glong _ABO;
    static gfloat ABO_offset = 0;
    gint c = 0;
    gboolean result = TRUE;
    gint samples;

    /* Init SCR */
    if (SCR==-1) {
        SCR = (gfloat)PTSvideo;
        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"SCR initialise to %f",SCR);
    }
    SCR33b = (glong)SCR;
    frame_delta = frame - frame_prev;
    _frame = (gfloat)frame_prev;
    /* Calc AUDIO_BYTES_PER_FRAME */
    if (AUDIO_BYTES_PER_FRAME==-1) {
        AUDIO_BYTES_PER_FRAME = drip_audio_samplerate() / drip_mpeg_framerate() * 4;
        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Output_audio: AUDIO_BYTES_PER_FRAME = %g",AUDIO_BYTES_PER_FRAME);
    }
    _ABO = ABO + (glong)(AUDIO_BYTES_PER_FRAME * frame_delta);

    /* Anything in video buffer? */
    if (encode_frame(NULL,NULL,FALSE,TRUE)==FALSE) {
        #ifdef DEBUG
        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Nothing in video buffer, canceling audio handling");
        #endif
        return;
    }

    /* Output audio till it's in sync with video */
    audio_buffer_filled = FALSE;
    audio_chunk = fetch_from_audio_buffer(&samples,SCR33b);
    while (audio_chunk!=NULL && result==TRUE) {
        avifile_audio_outstream->AddData(audio_chunk,samples/2); 
        avifile_audio_outstream_bytes += samples;
        ABO += samples;
        ABO_offset += samples;
        if (ABO_offset>(AUDIO_BYTES_PER_FRAME-1)) {
            ABO_offset -= AUDIO_BYTES_PER_FRAME;
            audio_buffer_filled = TRUE;
            result = encode_frame(stream,im,TRUE,FALSE);
        }
        if (result==TRUE) {
            SCR += PTS_PER_AUDIO_BYTE*samples;
            #ifdef DEBUG
            g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"ABO_offset = %g, samples = %i, AUDIO_BYTES_PER_FRAME = %g, avifile_audio_outstream_bytes = %li,  PTS++ = %g, SCR = %f\n",ABO_offset,samples,AUDIO_BYTES_PER_FRAME,avifile_audio_outstream_bytes,((gfloat)PTS_PER_AUDIO_FRAME)/1024*samples,SCR);
            #endif
            SCR33b = ((glong)SCR);// | 8589934592; // use 33bits of SCR
            audio_chunk = fetch_from_audio_buffer(&samples,SCR33b); 
        }
    }

    /* Update, clean & exit */
    frame_prev = (gint)_frame;
    return;
}


/* Handle audio blocks */
void a52_decode_data(guint8 *data, guint8 *end, gint PTS, gint PTSpos) {
    static a52_state_t state;
    static guint8 buf[65536];
    static guint8 *bufptr = buf;
    static guint8 *bufpos = buf + 7;
    static gfloat PTSa52 = -1;
    static guint8* start;
    static gint avifile_audio_total_offset = 0;
    /*
     * sample_rate and flags are static because this routine could
     * exit between the a52_syncinfo() and the ao_setup(), and we want
     * to have the same values when we get back !
     */
    static gint flags;
    gint pcm_size,samples;

    start = data;
    while (start<end) {
        *bufptr++ = *start++;
        if ((PTS!=-1) && (((glong)start-(glong)data)>PTSpos)) {
            // Only set the PTS once we've hit the magic pointer
            PTSa52 = PTS;
            PTS = -1;
        }

        if (bufptr == bufpos) {
            if (bufpos == buf + 7) {
                gint length;
                length = a52_syncinfo(buf,&flags,&audio_sample_rate,&audio_bit_rate);
                if (!length) {
                    for (bufptr=buf; bufptr<buf+6; bufptr++) {
                        bufptr[0] = bufptr[1];
                    }
                    continue;
                }
                bufpos = buf+length;
            } else {
                sample_t level,bias;
                gint i;

                flags = A52_ADJUST_LEVEL | A52_STEREO;
                bias = 384.0;
                level = 1.0;            // Volume adjustment, TODO less for 5.1?
                if (a52_frame(&state,buf,&flags,&level,bias)) {
                    g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"Audio decoder: Error in a52_frame");
                    goto error;
                }
                flags &= A52_CHANNEL_MASK;
                audio_channels = 2; // flags = A52_ADJUST_LEVEL | A52_STEREO
                // Decode audio 
                if (PTSa52>-1) {
                    for (i=0; i<6; i++) {
                        if (a52_block(&state,output_liba52)) {
                            g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Audio decoder: Error in a52_block");
                            goto error;
                        }
                        pcm_size = float_to_int_stereo(output_liba52,gint16_samples);
                        add_to_audio_buffer(gint16_samples,pcm_size,(gint)PTSa52);
                        PTSa52 += PTS_PER_AUDIO_FRAME;
                    }
                } 
                bufptr = buf;
                bufpos = buf + 7;
                continue;
            error:
                bufptr = buf;
                bufpos = buf + 7;
            }
        }
    }
    return;
}


/* Demux stream */
gpointer drip_demux(gpointer data, gint audio_channel) {
    static gint mpeg1_skip_table[16] = {
             1, 0xffff,      5,     10, 0xffff, 0xffff, 0xffff, 0xffff,
        0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff
    };
#define DEMUX_HEADER 0
#define DEMUX_DATA 1
#define DEMUX_SKIP 2
    static gint state = DEMUX_SKIP;
    static gint state_bytes = 0;
    guint8 *buf;
    size_t rbuf;
    guint8 *end;
    guint8 *tmp1;
    guint8 *tmp2;
    gint complain_loudly;
    gint len;
    guint headerlen, packetlen, streamid;
    guint a1, a2;
    glong PTS, DTS, SCR;
    gint pass = 0;
    glong bufstart = (glong)buf;
    gint bytes;
    static uint8_t head_buf[268];
    glong buf_previous = 0;

    complain_loudly = 1;
    buf = buffer;

    do {
        rbuf = fread(buf, 1, buffer + DEMUX_BUFFER_SIZE - buf, in_file);
        if (rbuf!=((glong)buffer+DEMUX_BUFFER_SIZE-(glong)buf) && input_files[input_files_index]!=NULL) {
            in_file = fopen(input_files[input_files_index],"rb");
            if (!in_file) g_log(DRIP_LD,G_LOG_LEVEL_ERROR,"%s - couldnt open file %s", strerror(errno),input_files[input_files_index]);
            else g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Opened %s for input",input_files[input_files_index]);
            input_files_index++;
            rbuf = fread(buf, 1, buffer + DEMUX_BUFFER_SIZE - buf, in_file);
        }
        end = buf + rbuf;
        buf = buffer;
        while ((buf + 4 <= end) /*&& (quit==FALSE)*/) {
            /* check start code */
            buf_previous = (glong)buf;
            if (buf[0] || buf[1] || (buf[2] != 0x01)) {
                if (complain_loudly) {
                    if ((buf[0] == 0) && (buf[1] == 0) && (buf[2] == 0))
                    complain_loudly = 0;
                }
                buf++;
                continue;
            }
            /* Get SCR */
            a1 = buf[4] << 24 | buf[5] << 16 | buf[6]  << 8 | buf[7];
            a2 = buf[8] << 24 | buf[9] << 16 | buf[10] << 8 | buf[11];
            SCR  = (a1 & 0x38000000) << 3;    // shifting the top bit out ..  . ...
            SCR |= (a1 & 0x03fff800) << 4;
            SCR |= (a1 & 0x000003ff) << 5;
            SCR |= (a2 & 0xf8000000) >> 27; 

            /* Get PTS or DTS */
            if( buf[7] & 0x80 ) { /* PTS */
                PTS  = (buf[ 9] & 0x0E) << 29 ;
                PTS |=  buf[10]         << 22 ;
                PTS |= (buf[11] & 0xFE) << 14 ;
                PTS |=  buf[12]         <<  7 ;
                PTS |= (buf[13] & 0xFE) >>  1 ;
            } else {
                PTS = 0;
            }
            if( buf[7] & 0x40 ) { /* DTS */
                DTS  = (buf[14] & 0x0E) << 29 ;
                DTS |=  buf[15]         << 22 ;
                DTS |= (buf[16] & 0xFE) << 14 ;
                DTS |=  buf[17]         <<  7 ;
                DTS |= (buf[18] & 0xFE) >>  1 ;
            } else {
                DTS = 0;
            }

            /* Handle id */
            switch (buf[3]) {
            case 0xb9:  /* program end code */
                return NULL;
            case 0xbb:  /* pss header */
                tmp1 = buf + (buf[4] << 8) + buf[5] + 5;
                buf = tmp1;
                break;
            case 0xba:  /* pack header */
                /* skip */
                if ((buf[4] & 0xc0) == 0x40) {   /* mpeg2 */
                    SCRstream = SCR;
                    tmp1 = buf + 14 + (buf[13] & 7);
                } else if ((buf[4] & 0xf0) == 0x20) {      /* mpeg1 */
                    SCRstream = SCR;  // TODO: SCR mpeg1 diff from mpeg2??
                    tmp1 = buf + 12;
                } else if (buf + 5 > end) {
                    goto copy;
                } else {
                    g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"Demuxer: Weird pack header");
                }
                if (tmp1 > end) {
                    goto copy;
                }
                buf = tmp1;
                break;
            case 0xe0:  /* video */
                tmp2 = buf + 6 + (buf[4] << 8) + buf[5];
                if (tmp2 > end) {
                    goto copy;
                }
                if ((buf[6] & 0xc0) == 0x80) {   /* mpeg2 */
                    tmp1 = buf + 9 + buf[8];
                } else {  /* mpeg1 */
                    for (tmp1 = buf + 6; *tmp1 == 0xff; tmp1++) {
                        if (tmp1 == buf + 6 + 16) {
                            buf = tmp2;
                            break;
                        }
                    }
                    if ((*tmp1 & 0xc0) == 0x40)
                        tmp1 += 2;
                    tmp1 += mpeg1_skip_table [*tmp1 >> 4];
                }
                if (tmp1 < tmp2) {
                    gint num_frames;
                    if (PTSvideo == -1) {
                        PTSvideo = PTS; //Store PTS of 1st frame
                        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Demuxer: 1st frame has PTS %i",PTSvideo);
                    }
                    if (PTS!=0) PTSframe = PTS; // store current PTS value, and use this one as PTS for next vide frame. Not accurate .. but untill libmpeg2 supplies PTS values with build frames this has todo;)
printf("PTSframe = %lu\n",PTSframe);
                    num_frames = mpeg2_decode_data(mpeg2dec, tmp1, tmp2);
                }
                buf = tmp2;
                break;
            case 0xbd: /* Private streams (SPU, Audio, others) */
                PTSprivate = PTS; /* store presentation time */
                len = buf[8];
                tmp1 = (guint8*)((glong)buf + 6 + (buf[4]<<8) + buf[5]);
                spu_track = buf[len+9] & 0x1F;
                audio_track = buf[len+9] & 0x0F;
                // Subpictures
                if ((buf[len+9] & 0xE0) == 0x20) {
                    spu_render(buf-14);
                }
                // Audio
                if (((buf[len+9] & 0xF0)==0x80) && (audio_track==audio_channel)) {
                    /* A52 audio */
                    if(audio_channel>=0) {
                        a52_decode_data((guint8*)((glong)buf+len+13),(guint8*)((glong)buf+len+2029),PTSprivate,buf[len+11]<<8|buf[len+12]);
                    }
                } else if (((buf[len+9] & 0xF0)==0xA0) || (audio_track==(audio_channel-16))) {
                    /* PCM audio */
                    gint pcm_offset;
                    gint offset = (buf[len+10]<<8) + buf[len+11];
                    const gfloat PTSsub_const = (90000.0/48000.0)/4;
                    gint PTSsub = (gint)((double)(offset)*PTSsub_const);
                    PTSprivate = PTS - PTSsub;
                    packetlen = (buf[4]<<8) + buf[5] - 3 - buf[8];
                    for (pcm_offset=0; ++pcm_offset<packetlen-1; ) {
                        if (buf[len+9+pcm_offset]==0x01 && buf[len+9+pcm_offset+1]==0x80 ) {
                            /* Start */
                            pcm_offset += 2;
                            break;
                        }
                    }
                    if (audio_channel>=0) {
                        guint8* audio_data = (guint8*)((glong)&buf[len+9+pcm_offset]);
                        guint8 tmp;
                        for (gint i=0;i<(packetlen-pcm_offset);i+=2) {
                           tmp = audio_data[i];
                           audio_data[i] = audio_data[i+1];
                           audio_data[i+1] = tmp;
                        }
                        if (PTS) {
                            add_to_audio_buffer((gint16*)(audio_data),packetlen-pcm_offset,PTSprivate);
                        } else {
                            g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"Demuxer: no PTS on PCM audio frame");
                        }
                    }
                }
                buf = tmp1;
                break;
            default:
                if (buf[3] < 0xb9) {
                    g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"Demuxer: looks like a video stream, not system stream");
                }
                /* skip */
                tmp1 = buf + 6 + (buf[4] << 8) + buf[5];
                if (tmp1 > end)
                    goto copy;
                buf = tmp1;
                break;
            }
        }
        if (buf < end) {
        copy:
            /* we only pass here for mpeg1 ps streams */
            memmove (buffer, buf, end - buf);
        }
        buf = buffer + (end - buf);
    } while ( (end == buffer+DEMUX_BUFFER_SIZE) && (quit==FALSE) );
    demuxing = FALSE;

    /* Close libraries */
    mpeg2_close(mpeg2dec);
    vo_close(video_output);
    free(mpeg2dec);
    fclose(in_file);
    free(output_liba52);
    ao_close(audio_output);
    return NULL;
}


/* Libmpeg2 has a new frame ready */
static void drip_draw_frame(vo_frame_t *frame) {
    total_frames++; // track total frames build
    drip_framerate(mpeg2dec); // set framerate to current
    if (quit==TRUE) return;
    if (processor!=NULL) {
        processor(frame->base); // Call processor [set in drip_mpeg_init()]
    } else {
        g_log(DRIP_LD,G_LOG_LEVEL_ERROR,"Cant process frame, this is a bug");
    }
    return;
}


/* Start demuxing dvd stream */
void drip_mpeg_start_demuxing(gint audio_channel) {
    /* Start (threadless?) demuxing */
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Starting demux thread");
    demuxing = TRUE;
    process = TRUE;
    quit = FALSE;
    drip_demux(NULL,audio_channel);
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Ended demux thread");
    return;
}


/* Initialse libvo output */
static gint drip_null_setup(vo_instance_t *instance, gint w, gint h) {
    gint result;
    // publice w & h
    width = w;
    height = h;
    result = libvo_common_alloc_frames(instance,width,height,sizeof(null_frame_t),NULL,NULL,drip_draw_frame);
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"drip_null_setup done (%i)",result);
    return result;
}


/* Open libvo output */
vo_instance_t* drip_vo_null_open(void) {
    gint c;
    vo_driver_t *drivers;
    vo_open_t *output_open = NULL;
    gint i;
    null_instance_t* instance;
    const gchar* driver_name = "drip";

    if (strcmp(driver_name,"drip")==0) {
        instance = (null_instance_t*)malloc(sizeof(null_instance_t));
        if (instance==NULL) return NULL;
        instance->vo.setup = drip_null_setup;
        instance->vo.close = libvo_common_free_frames;
        instance->vo.get_frame = libvo_common_get_frame;
        vo_open_instance = (common_instance_t*)instance;
        return (vo_instance_t*)instance;
    } else {
        drivers = vo_drivers();
        for (i=0; drivers[i].name != NULL; i++) {
            printf("driver %s\n",drivers[i].name);
            if (strcmp (drivers[i].name, driver_name) == 0) {
                output_open = drivers[i].open;
            }
        }
        return (vo_instance_t*)output_open;
    }
}


/* Returns NULL audio driver of libao */
ao_open_t *drip_audio_null_driver(void) {
    ao_driver_t *drivers;
    ao_open_t *driver;

    drivers = ao_drivers (); // Get drivers list
    for (gint i=0; drivers[i].name!=NULL; i++) {
        if (strcmp(drivers[i].name, "null") == 0) {
             driver = drivers[i].open;
        }
    }
    if (driver==NULL) g_log(DRIP_LD,G_LOG_LEVEL_ERROR,"Could not open audio output");
    return driver;
}


/* Opens libmpeg2 handlers */
gint drip_mpeg_open(gchar *input,void handler(guint8*[3])) {
    gint i;
    // open input file
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Opening input mpeg from %s",input);
    input_files = g_strsplit(input,":",256);
    input_files_index = 0;
    in_file = fopen(input_files[input_files_index],"rb");
    if (!in_file) g_log(DRIP_LD,G_LOG_LEVEL_ERROR,"MPEG Open: %s - couldnt open file %s", strerror(errno),input_files[input_files_index]);
    else g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Opened %s for input",input_files[input_files_index]);
    input_files_index++;

    // process handler
    processor = handler;
    // open libmpeg2 output
    video_output = vo_open(drip_vo_null_open);
    // init libmpeg2 (video)
    mpeg2dec = (mpeg2dec_t*)malloc(sizeof(mpeg2dec_t));
    memset(mpeg2dec,0,sizeof(mpeg2dec_t));
    mpeg2_init(mpeg2dec,mm_accel()|MM_ACCEL_MLIB,video_output);

    // open liba52 output
    audio_output_open = drip_audio_null_driver();
    audio_output = ao_open(audio_output_open);
    memset(silence,256*2*sizeof(gint16),0);
    // init liba52 (audio)
    output_liba52 = a52_init(mm_accel()|MM_ACCEL_MLIB);
    if (output_liba52==NULL) g_log(DRIP_LD,G_LOG_LEVEL_ERROR,"Failed to initialize audio");
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Initialized import of mpeg video and audio");
    return TRUE;
}


/* Closes libmpeg2 handlers */
gint drip_mpeg_close(void) {
    quit = TRUE;
    g_strfreev(input_files);
    reset_audio_buffer();
    return TRUE;
}

