// Jarl van Katwijk - 2001
//                    2002 februari : mayor update, linked to libmpeg2 and added fifo's
// 
// original based on mpeg2divx 0.8
// 2000/10/23 Ulrich Hecht <uli@emulinks.de>
// licensed under the GNU Public License (GPL) v2
//
// added second audio stream - Rainer Lay 2001

/* system */
#include <sys/time.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <glib.h>
#include <gtk/gtk.h>
#include <unistd.h>
#include <fstream>
#include <errno.h>
#include <pthread.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
/* mpeg3 decoding */
#include "demuxer.hh"
/* drip general */
#include "../src/drip.h"
#include "../src/external.h"
/* filters */
#include "deinterlace.hh"
#include "autoclipper.hh"
#include "stablelizer.hh"
#include "pulldown.hh"
#include "scaler.hh"
#include "clipper.hh"
#include "conversion.hh"
#include "plugin-loader.hh"
/* main */
#include "main.hh"
#include "fast_memcpy.hh"

using namespace std;

// maximum number of audio streams
#define AMAX 2
//#define DEBUG

// protos
void set_dimensions_aspect(void);
void allocate_buffers(guchar* buffer);

// types
typedef struct {
    gint in_stream;
    fourcc_t codec;
    gint mp3bitrate;
    int achan;
    int rate;
    long samples;
    gfloat sampsperframe_add;
    gint sampsperframe;
    gfloat samplepos;
    gfloat samplepos_last;
    gshort* audiobuffer;
    WAVEFORMATEX fmt;
    IAviAudioWriteStream* outstream;
    gint volume_adjustment;
} au_t;
 
typedef struct {
    gint rc_reaction_period;
    gint rc_reaction_ratio;
    gint rc_period;
    gint max_key_interval;
    gint obmc;
    gint quality;
    gint max_quantizer;
    gint min_quantizer;
} opendivx_params_t;

// globals   TODO: a little much .. ;(
IAviAudioWriteStream* avifile_audio_outstream;
gfloat aspect = 0;
gboolean scale = FALSE;
guint width;
guint height;
gint owidth=-1;
gint oheight=-1;
gfloat framerate;
fourcc_t video_codec=fccDIV3;
BitmapInfo bminfo_in; // Codec struct
CImage* im; // framebuffer image 
CImage* progress_frame_im; // I420->RGB conversion buffer
gint clipleft = 0;
gint clipright = 0;
gint cliptop = 0;
gint clipbottom = 0;
gint max_size = 731906048;
glong frames;
au_t au[AMAX];
gboolean autoclip = FALSE;
guchar* framebuffer;
guint8 *clippedbuffer[3];
guchar ToRGBbuffer[720*576*3];
u_int8_t *planes[3];
guchar* extrabuffer = NULL; // ! extrabuffer == NULL -> no setup done yet
guchar** rowptr;
guint8* scaledbuffer[3];
guchar* _RGBbuffer;
gchar* outputfile;
gint divxbitrate=1200;
gint subpicture = -1;
gchar* clutfile = NULL;
gboolean pulldown = FALSE;
GdkPixbuf* gframe;
GdkPixbuf* gscaled;
opendivx_params_t opendivx_params;
gint endframe=-1;
GString* inputfiles[256];
IAviSegWriteFile* avifile;
IAviVideoWriteStream* stream;
glong todo_frames;
gint quit;
gboolean audio_encoding;
gint numfiles;
gint file;
gint result;
GString *output;
gint argbase=0;
gfloat mp3bitrate=-1;
gboolean deinterlace=FALSE;
gint austream=0;
gint vistream=0;
gint limit=0;
gboolean moreopts=true;
struct stat VOB_file;
gulong total_bytes;
gfloat frames_per_bytes;
glong totalframes=0;
struct timeval tv;
struct timezone tz={0,0};
glong startsec;
glong startusec;
double fps=0;
gint total_channels;
gint assetChannels;
int16_t *input_ptr;
int16_t *output_ptr;
gchar *buffer;
// audio
gchar *audio_buffer;
gint downmix51;
int16_t *input_end;
off_t len;
gint cached_frames = 0;
glong totalframes_countdown = 0;
GString *framefile;
glong writeout_frames_countdown = 10;
glong writeout_frames_countdown_amount = 10;
gboolean audio_output_codec_setup = FALSE;
gboolean audio_buffer_filled;
/* Frame buffer */
typedef struct _plane_t plane_t;
struct _plane_t {
    guint8 *planes[3];
};
static gint frame_buffer_index = -1;
static gint frame_buffer_base = 0;
const gint frame_buffer_index_MAX = 128;
static plane_t frame_buffer[frame_buffer_index_MAX];
static guint8 *frame_raw = NULL;
enum {NONE,CODEC,BUFFER,FETCH,ALL};
static gboolean first_frame = TRUE;
guint8* f_buffer_bak[3];
gboolean noise = FALSE;
guint8 deinterlace_buffer[720*576];
gfloat SCRdrip;  // Drip's System Clock Referance value
gfloat SCRframe; // Value to increase SCRdrip every video frame


/* Wrapper for SetAttribute of avifile codec object */
void Codec_SetAttribute(const CodecInfo* codecInfo,const gchar* attribute, gint value) {
    gint result;
    #ifndef STANDALONE
    /* Lock */
    drip_lock("Codec SetAttribute");
    #endif
    /* Try to set attribute */
    result = Creators::SetCodecAttr(*codecInfo, attribute, value);
    #ifndef STANDALONE
    /* Unlock */
    drip_unlock("Codec SetAttribute");
    #endif
    /* Handle negative sets */
    if (result==-1) {
        g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"Could not set attribute %s to value %i. Ignoring, wrong codec or outdated avifile?",attribute,value);
    }
    return;
}


/* Setup audio output codec */
void audio_output_codec(void) {
    /* Check if audio is already available */
    /* NOTE: untill audio_output_codec_setup is TRUE chunks
             of silence will be added to audio channel */
    if ((audio_output_codec_setup == TRUE)) {
        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"audio_output_codec, audio already setup");
        return;
    }
    /* Output audio codec */
    switch (au[0].codec) {
        case 85:         g_log(DRIP_LD,G_LOG_LEVEL_MESSAGE,"Encoding audio to MP3");
                         break;
        default:         g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"No audio codec, not encoding.");
                         return;
    }
    /* Parse audio info */
    audio_encoding=FALSE;
    for (gint anum=0;anum<AMAX;anum++) {
        if (au[anum].in_stream>-1) {
            au[anum].achan = 2;//drip_audio_channels(); //TODO
            au[anum].rate  = drip_audio_samplerate();
            audio_encoding = TRUE;
            g_log(DRIP_LD,G_LOG_LEVEL_MESSAGE,"Audio %d: Audio channels %d,rate %d",anum,au[anum].achan,au[anum].rate);
        }
    }
    /* Audio channels */
    if (!audio_encoding) 
        g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"There will be NO audio encoded!");
    /* Output Audio header */
    for (gint anum=0;anum<AMAX;anum++) {
        if (au[anum].in_stream>-1) {
            /* Output Audio header */
            memset(&au[anum].fmt,0,sizeof(WAVEFORMATEX));
            au[anum].fmt.wFormatTag = 1; /* PCM audio id */
            au[anum].fmt.nChannels = au[anum].achan;
            au[anum].fmt.nSamplesPerSec = au[anum].rate;
            au[anum].fmt.wBitsPerSample = 16;
            au[anum].fmt.nBlockAlign = 2;//au[anum].fmt.wBitsPerSample/8 * au[anum].fmt.nChannels;//2
            au[anum].fmt.nAvgBytesPerSec = au[anum].fmt.nSamplesPerSec;// * au[anum].fmt.nBlockAlign;
            au[anum].fmt.cbSize = sizeof(WAVEFORMATEX);
            au[anum].sampsperframe_add = au[anum].rate / drip_mpeg_framerate();//au[anum].fmt.nAvgBytesPerSec / drip_mpeg_framerate();
            au[anum].sampsperframe = (gint)au[anum].sampsperframe_add + 1;
            au[anum].samplepos = 0;
            au[anum].samplepos_last = 0;
            if (au[anum].mp3bitrate == -1) {
                switch(au[anum].rate) {
                case 48000: au[anum].mp3bitrate=(au[anum].achan>1?160:80*au[anum].achan); break;
                case 44100: au[anum].mp3bitrate=(au[anum].achan>1?112:56*au[anum].achan); break;
                case 22050: au[anum].mp3bitrate=(au[anum].achan>1?64:32*au[anum].achan);break;
                default:    g_log(DRIP_LD,G_LOG_LEVEL_ERROR,"Audio %d: Unknown sample rate %d Hz, specify MP3 bitrate",anum,au[anum].rate);
                            exit(1);
                } 
            }
            if (au[anum].achan>0) {
                #ifndef STANDALONE
                drip_lock("audio_output_codec");
                #endif
                au[anum].outstream=avifile->AddAudioStream(au[anum].codec,&au[anum].fmt,(gint)(au[anum].mp3bitrate*1000)/8);
                avifile_audio_outstream = au[anum].outstream;
                au[anum].outstream->Start();
                #ifndef STANDALONE
                drip_unlock("audio_output_codec");
                #endif
                g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Audio %d: MP3 bitrate %ibps",anum,au[anum].mp3bitrate);
            }
        }
    }
    audio_output_codec_setup = TRUE;
    return;
}


/* Setup video output codec */
void video_output_codec(void) {
    /* Video header */
    bminfo_in = BitmapInfo(owidth,oheight,AVIFILE_COLOUR_DEPT_OUTPUT);
    bminfo_in.SetSpace(fccI420);
    /* Various encoding settings */
    gint quality = 10000;
    gint keyfreq = 200;
    /* Output codec */
    switch (video_codec) {
        #if (AVIFILE_MAJOR_VERSION == 0) && (AVIFILE_MINOR_VERSION==6) 
        case fccDIVX:    g_log(DRIP_LD,G_LOG_LEVEL_MESSAGE,"Encoding video to OpenDIVX");
                         break;
        #endif
        case fccDIV3:    g_log(DRIP_LD,G_LOG_LEVEL_MESSAGE,"Encoding video to MPEG4 low motion");                         break;
        case fccDIV4:    g_log(DRIP_LD,G_LOG_LEVEL_MESSAGE,"Encoding video to MPEG4 high motion");
                         break;
        default:         g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"No video codec, not encoding.");
                         return;
    }
    /* Output codec attributes */
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"About to initialise avifile, when Drip stops (crashes) here get yourself a working avifile...");
    /* Create output avi stream */
    avifile = CreateSegmentedFile(outputfile,max_size,0,00777);
    #if (AVIFILE_MAJOR_VERSION == 0) && (AVIFILE_MINOR_VERSION==6) 
    const CodecInfo* codecInfo=CodecInfo::match(video_codec);
    if (video_codec==fccDIV3 || video_codec==fccDIV4) { /* MSMPEG4 */
       Codec_SetAttribute(codecInfo, "Crispness", 100);
       Codec_SetAttribute(codecInfo, "KeyFrames", keyfreq);
       Codec_SetAttribute(codecInfo, "BitRate", divxbitrate);
    }
    if (video_codec==fccDIVX) { /* OpenDIVX, DIVX4 */
        Codec_SetAttribute(codecInfo, "BitRate",divxbitrate*1500);
        Codec_SetAttribute(codecInfo, "rc_period",keyfreq*10);
        //What's this? : Codec_SetAttribute(codecInfo, "MaxAuto",6);
        Codec_SetAttribute(codecInfo, "rc_reaction_period",(gint)(keyfreq/20));
        Codec_SetAttribute(codecInfo, "rc_reaction_ratio",(gint)(keyfreq/10));
        Codec_SetAttribute(codecInfo, "max_key_interval",(gint)((keyfreq*3)/2));
        Codec_SetAttribute(codecInfo, "obmc",0); //TODO
        Codec_SetAttribute(codecInfo, "quality",5); //opendivx_params.quality); // DIVX:5=best,1=lowest, avifile:1=enable VBR... ????
        Codec_SetAttribute(codecInfo, "max_quantizer",opendivx_params.max_quantizer);
        Codec_SetAttribute(codecInfo, "min_quantizer",opendivx_params.min_quantizer);
        Codec_SetAttribute(codecInfo, "bidirect",0); //TODO:2pass encoding,0 = disabled
        Codec_SetAttribute(codecInfo, "deinterlace",0); //TODO
    }
    #else
    IVideoEncoder::SetExtendedAttr(codec, "BitRate", divxbitrate);
    IVideoEncoder::SetExtendedAttr(codec, "KeyFrames", keyfreq);
    IVideoEncoder::SetExtendedAttr(codec, "Crispness", 100);
    #endif
    /* Add video stream to output avi */
    #ifndef STANDALONE
    drip_lock("video_output_codec");
    #endif
    stream = avifile->AddVideoStream(video_codec, &(bminfo_in), (guint)(1000000./framerate));
    stream->SetQuality(quality);
    stream->SetKeyFrame(keyfreq);
    stream->Start();
    #ifndef STANDALONE
    drip_unlock("video_output_codec");
    #endif
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Avifile initialisation calls passed");
    g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Video Encoding bitrate %i",divxbitrate);
    g_log(DRIP_LD,G_LOG_LEVEL_INFO,"DIVX Chunk size %i bytes",max_size);
    g_log(DRIP_LD,G_LOG_LEVEL_MESSAGE,"DIVX Output File %s",outputfile);
    /* TEST */
    audio_output_codec();
    return;
}


/* Autoclipping */
void autoclipping(void) {
    gint samples = 5; // TODO!
    gint factor = 16;
    /* Anything todo? */
    if (autoclip==FALSE) return;
    if (endframe<200) {
        g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Autoclipper disabled because not enough frames (200) in source");
        autoclip = FALSE;
    }
    /* Set clip values to high */
    cliptop = 1000;clipbottom = 1000;clipleft = 1000;clipright = 1000;
    /* Run autoclipper */
    autoclipper_apply(inputfiles[0]->str,endframe,1,samples);
    /* Round clipping values to 4 */
    factor = 2;
    cliptop    = ((gint)((gfloat)cliptop/factor)*factor);
    clipbottom = ((gint)((gfloat)clipbottom/factor)*factor);
    clipleft   = ((gint)((gfloat)clipleft/factor)*factor);
    clipright  = ((gint)((gfloat)clipright/factor)*factor);
    /* Any clipping -> set scale to TRUE */
    if ((cliptop!=0)||(clipbottom!=0)||(clipleft!=0)||(clipright!=0)) {
        scale = TRUE;
    };
    width = drip_mpeg_width();
    height = drip_mpeg_height();
    /* Complete clipping? */
    if ((cliptop>height) || (clipbottom>height) || (clipleft>width) || (clipright>width)) {
        g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Autoclipping disabled, empty or too less video");
        cliptop    = 0;
        clipbottom = 0;
        clipleft   = 0;
        clipright  = 0;
    } else {
        g_log(DRIP_LD,G_LOG_LEVEL_MESSAGE,"Clipping top %i,bottom %i,left %i,right %i",cliptop,clipbottom,clipleft,clipright);
        g_log(DRIP_LD,G_LOG_LEVEL_MESSAGE,"Input video clipped %ix%i",width-(clipleft+clipright),height-(cliptop+clipbottom));
    }
    return;
}

/* Clip frame */
void clip_frame_i420(guint8 *src[3], guint8 *dest[3],gint xoff,gint yoff,gint w, gint h,gint w_in,gint h_in) {
    clipper_apply(src,dest,xoff,yoff,w,h,w_in,h_in);
    return;
}

/* Scale I420 frame */
void scale_frame_i420(guint8* source[3], guchar* dest,gint swidth,gint sheight,gint dwidth,gint dheight, gint cleft, gint cright, gint ctop, gint cbottom) {
    scaler_apply(swidth,sheight,dwidth,dheight,source,dest);
    return;
}


/* Encode frame */
gboolean encode_frame(IAviVideoWriteStream *stream,CImage *frame,gboolean fetching,gboolean probing) {
    gint i,amount;
    gint writeout_frame = NONE;
    gboolean result = TRUE;
    static guint WH32 = owidth*oheight*3/2;

    /* Allocate buffer */
    if (frame_raw == NULL) {
        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"FB: Allocating frame buffer for %ix%i",owidth,oheight);
        for (i=0;i<frame_buffer_index_MAX;i++) {
            frame_buffer[i].planes[0] = (guint8*)malloc(WH32);
        }
        /* Locate raw buffer */
        if (scale == TRUE) {
            frame_raw = (guchar*)scaledbuffer[0];
        } else {
            frame_raw = framebuffer;
        }
        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"FB: Allocated frame buffer");
    }
    /* Only probing buffer? */
    if (probing==TRUE) {
        if (frame_buffer_index>0) {
            return TRUE;
        } else {
            return FALSE;
        }
    }
    /* Fetching frame from buffer? */
    if (fetching==TRUE) {
        if (audio_buffer_filled==TRUE) {
            writeout_frame = FETCH;
        }
    } else {
        writeout_frame = BUFFER;
    }

    /* Execute action */
    switch (writeout_frame) {
        case CODEC:  /* Feed frame & audio to codec */
                     amount = stablelizer_video(PTSframe);
                     for (i=0;i<amount;i++) {stream->AddFrame(frame);}
                     /* Frame counter */
                     totalframes++;
                     break;
        case BUFFER: if (first_frame == TRUE) {
                         /* DONT use very 1st frame */
                         first_frame = FALSE;
                         break;
                     }
                     if ((frame_buffer_index+frame_buffer_base)>=frame_buffer_index_MAX) {
                         for (i=0;i<frame_buffer_index;i++) {
                             fast_memcpy(frame_buffer[i].planes[0],frame_buffer[i+frame_buffer_base].planes[0],WH32);
                         }
                         frame_buffer_base = 0; //Reset base
                     }
                     if (frame_buffer_index<frame_buffer_index_MAX) {
                         /* Add new frame to buffer */
                         #ifdef DEBUG
                         g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"FB: Buffering, copying %li bytes (w=%i,h=%i)  index=%i, base=%i, frame=%li",WH32,owidth,oheight,frame_buffer_index,frame_buffer_base,totalframes);
                         #endif
                         fast_memcpy(frame_buffer[frame_buffer_index+frame_buffer_base].planes[0],frame_raw,WH32);
                         frame_buffer_index++;
                     } else {
                         result = FALSE;
                         g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"FB: Cant add new frame to frame buffer");
                     }
                     break;
        case FETCH:  if (frame_buffer_index>0) {
                         /* Feed frame from fifo buffer to codec */
                         fast_memcpy(frame_raw,frame_buffer[frame_buffer_base].planes[0],WH32);
                         amount = stablelizer_video(PTSframe);
                         for (i=0;i<amount;i++) {stream->AddFrame(frame);}
                         /* Frame counter */
                         totalframes++;
                         /* Update base & index */
                         frame_buffer_base++;
                         frame_buffer_index--;
                         /* Rotate buffer */
                         if ((frame_buffer_index+frame_buffer_base)>=frame_buffer_index_MAX) {
                             for (i=0;i<frame_buffer_index;i++) {
                                 fast_memcpy(frame_buffer[i].planes[0],frame_buffer[i+frame_buffer_base].planes[0],WH32);
                             }
                             frame_buffer_base = 0; //Reset base
                             #ifdef DEBUG
                             g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"FB: Rotated frame buffer");
                             #endif
                         }
                     } else {
                         result = FALSE;
                         g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"FB: No frame in buffer to fetch");
                     }
                     break;
        case ALL:    if (frame_buffer_index<frame_buffer_index_MAX) {
                         /* Add new frame to buffer */
                         frame_buffer[frame_buffer_index].planes[0] = (guint8*)malloc(WH32);
                         fast_memcpy(frame_buffer[frame_buffer_index].planes[0],frame_raw,WH32);
                         frame_buffer_index++;
                         /* Feed frame from fifo buffer to codec */
                         fast_memcpy(frame_raw,frame_buffer[0].planes[0],WH32);
                         amount = stablelizer_video(PTSframe);
                         for (i=0;i<amount;i++) {stream->AddFrame(frame);}
                         /* Frame counter */
                         totalframes++;
                         /* Rotate buffer */
                         for (i=0;i<frame_buffer_index-1;i++) {
                             fast_memcpy(frame_buffer[i].planes[0],frame_buffer[i+1].planes[0],WH32);
                         }
                         frame_buffer_index--;
                         free(frame_buffer[frame_buffer_index].planes[0]);
                         break;
                     } else result = FALSE;
        default:     result = FALSE;
                     g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"FB: frame buffer UNKNOWN action, index=%i, frame=%li",frame_buffer_index,totalframes);
                     break;
    }
    /* Clean & Exit */
    if (result==FALSE) g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"FB: Error occured, index=%i, frame=%li",frame_buffer_index,totalframes);
    return result;
}


/* Allocate buffers */
void allocate_buffers(void) {
    framebuffer = (guchar*)malloc(720*576*3);
    /* Allocate video buffers */
    extrabuffer = (guchar*)malloc(width*BYTES_PER_COLOUR_OUTPOUT+4);
    rowptr = new (guchar*)[height];
    for (guint i=0; i<height; i++) {
        if(i==height-1) {
            rowptr[i]=extrabuffer;
        } else {
            rowptr[i]=framebuffer+(width*(height-1)*BYTES_PER_COLOUR_OUTPOUT)-(width*BYTES_PER_COLOUR_OUTPOUT*i);
        }
    }
    /* Fill colour planes pointers */
    planes[0] = framebuffer;
    planes[1] = planes[0] + width * height;
    planes[2] = planes[1] + (width * height)/4;
    /* Alloc clipped buffer */
    clippedbuffer[0] = (guint8*)malloc(720*576);
    clippedbuffer[1] = (guint8*)malloc(720*576/4);
    clippedbuffer[2] = (guint8*)malloc(720*576/4);
    scaledbuffer[0] = new unsigned char[labs(bminfo_in.biHeight*bminfo_in.biWidth*BYTES_PER_COLOUR_OUTPOUT)];
    scaledbuffer[1] = (guint8*)((glong)scaledbuffer[0]+owidth*oheight);
    scaledbuffer[2] = (guint8*)((glong)scaledbuffer[1]+owidth*oheight/4);
    _RGBbuffer = (guchar*)malloc(labs(bminfo_in.biHeight*bminfo_in.biWidth*3));
    im = new CImage(&bminfo_in,scaledbuffer[0], false);
    progress_frame_im =  new CImage(&bminfo_in,ToRGBbuffer,false);
    /* Bakup f_buffer */
    f_buffer_bak[0] = (guint8*)malloc(width*height);
    f_buffer_bak[1] = (guint8*)malloc(width*height/4);
    f_buffer_bak[2] = (guint8*)malloc(width*height/4);
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Allocated buffers for %ix%i frame (%ix%i)",width,height,bminfo_in.biWidth,bminfo_in.biHeight);
    return;
}

/* Set video dimensions to requested based on input size, and adjust for aspect */
void set_dimensions_aspect(void) {
    frames=-1;//frames=mpeg3_video_frames(file, vistream);
    g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Input video reads %dx%d %.2ffps",width,height,framerate);
    /* Automatic width\height */
    gboolean correct_aspect = FALSE;
    if(oheight==-1 && owidth==-1) { // All auto
        owidth=width-(clipleft+clipright);
        oheight=height-(cliptop+clipbottom);
        correct_aspect = TRUE;
    }
    if(owidth==-1 && oheight!=-1) { // Width auto
        owidth=width-(clipleft+clipright);
        correct_aspect = TRUE;
    }
    if(oheight==-1 && owidth!=-1) { // Height auto
        oheight = height-(cliptop+clipbottom);
        correct_aspect = TRUE;
    }
    /* Aspect scaling */
    if (correct_aspect==TRUE) {
        gfloat ow = owidth;
        gfloat oh = oheight;
        gfloat as = (ow/oh);
        gfloat correct;
        if (aspect==0) {
            /* No aspect given, correct based on sizes we got so far, no very correct... */
            if (as>2.06388888) {
                /* 2.35:1 aspect */
                aspect = (gfloat)2.35/(gfloat)1;
            }
            if ((as>1.5555555) && (as<2.06388889)) {
                /* 16:9 aspect */
                aspect = (gfloat)16/(gfloat)9;
            }
            if (as<1.5555556) {
                /* 4:3 aspect */
                aspect = (gfloat)4/(gfloat)3;
            }
        }
        g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Using DVD aspect %.3f:1",aspect);
        correct = aspect / ((gfloat)width/(gfloat)height);
        oheight = (gint)(oh / correct);
        oheight = (gint)(oheight / ((gfloat)width/(gfloat)owidth));
    }
    /* Adjust Width & Height to a multipe of 16 for better encoding */
    owidth = ((gint)owidth/16)*16;
    oheight= ((gint)oheight/16)*16;
    if ((oheight!=height)||(owidth!=width)) {  /* Scaling needed? */
        scale = TRUE;
    } else {
        scale = FALSE;
    }
    g_log(DRIP_LD,G_LOG_LEVEL_MESSAGE,"Output video corrected %ix%i at %.2fFPS",owidth,oheight,framerate);
    return;
}


/* Display dripencoder CLI mode usage and exit() */
void display_help(gchar* progname) {
    printf("\nUsage: %s [OPTION]... [input MPEG]... [output AVI]\n",progname);
    printf("\nOptions:\n");
    printf("  -b\tDivX ;-) bitrate (default 1200)\n");
    printf("  -f\tuse fast-motion codec (default low-motion)\n");
    #if (AVIFILE_MAJOR_VERSION == 0) && (AVIFILE_MINOR_VERSION==6) 
    printf("  -d\tuse DIVX4 encoding codec\n");
    printf("  -dq\tDIVX4 quality\n");
    printf("  -dminq\tDIVX4 min quantizer\n");
    printf("  -dmaxq\tDIVX4 max quantizer\n");
    printf("  -c\tuse avifile video codec (for use with Drip GUI)\n");
    printf("  -au\tuse avifile audio codec (for use with Drip GUI)\n");
    #endif
    if (AMAX>1)
        printf("  -au2\tuse avifile audio codec (for use with Drip GUI) for audio channel 2\n");
    printf("  -a\tMP3 bitrate (kBit, default auto)\n");
    if (AMAX>1)
       printf("  -a2\tMP3 bitrate (kBit, default auto) for audio channel 2\n");
    printf("  -av\tAudio volume normalisation (0=off, 1=on)\n");
    if (AMAX>1)
       printf("  -av2\tAudio volume normalisation for audio channel 2\n");
    printf("  -e\tnumber of frames to encode (default all)\n");
    printf("  -ct\tnumber of lines to clip from top (default none)\n");
    printf("  -cb\tnumber of lines to clip from bottom (default none)\n");
    printf("  -cl\tnumber of lines to clip from left (default none)\n");
    printf("  -cr\tnumber of lines to clip from right (default none)\n");
    printf("  -ac\tauto clip (default off, disables other clipping options)\n");
    printf("  -w\toutput width (-1 = input width)\n");
    printf("  -h\toutput height (-1 = auto, when width=-1 too both are same as input)\n");
    printf("  -s\tadd subpictures to result, also needs a CLUT file\n");
    printf("  -r\taspect ratio\n");
    printf("  -ip\tdeinterlace PAL video stream (default off)\n");
    //printf("  -in\tdeinterlace (reverse 3:2 pulldown) NTSC video stream (default off) !! Not yet working !!\n");
    printf("  -as\taudio stream number (default 0)\n");
    if (AMAX>1)
        printf("  -as2\taudio stream number (default disabled) for audio channel 2\n");
    printf("  -vs\tvideo stream number (default 0)\n");
    printf("  -l\tlimit output to size in kb, open new output when limit is reached\n");
    printf("\nNOTE: input files should be standalone MPEG2 files, dvd .VOB files\n");
    printf("      and cached drip files are not valid mpeg2 files and must be read\n");
    printf("      in concationation mode: file1.vob:file2.vob:file3.vob\n\n");
    printf("\n" APPNAME " " VERSION " by Jarl van Katwijk <jarl@xs4all.nl>, based on work by Ulrich Hecht <uli@emulinks.de>\n");
    exit(0);
    return;
}


/* Setup encoding environment */
void init_environment(void) {
    /* Fill video stream values */
    width     = drip_mpeg_width();
    height    = drip_mpeg_height();
    framerate = drip_mpeg_framerate();
    /* Init filters */
    set_dimensions_aspect();
    video_output_codec();
    allocate_buffers();
    /* Init scaler */
    scaler_init(width-(clipleft+clipright),height-(cliptop+clipbottom),owidth,oheight);
    /* Init subpicture */
    if (subpicture>-1) {
        #ifndef STANDALONE
        drip_lock("spu_init");
        #endif
        spu_init(subpicture,width,height,framerate,clutfile,DRIP_CB_LD);
        #ifndef STANDALONE
        drip_unlock("spu_init");
        #endif
        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Initialized subpicture %i",subpicture);
    }
    /* Init reverse 3:2 pulldown */
    if (pulldown) {
        pulldown_init(width,height,false,0);  //TODO: odd frame & offset should not be hardcoded
        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Initialized reverse 3:2 pulldown");
    }
    #ifndef STANDALONE
    I420toRGB24_init();
    #endif
    /* Init stablelizer */
    stablelizer_init(framerate,drip_audio_samplerate(),width,height);
    /* Initialise plugins */
    for (gint i=0;i<filters_index;i++) {
        (*filters[i].module_init_function)(owidth,oheight,0,0,owidth,oheight,framerate,f_buffer_bak,f_buffer_bak);
    }
    /* Define SCRframe (value by which SCRdrip is increased every video frame */
    SCRframe = 90000 / drip_mpeg_framerate();
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Encoding environment setup");
}


/* Apply all filters and encode a frame with avifile */
void process_frame(guint8 **f_buffer) {
    guint8* current_buffer_planar[3];  // pointers to most recent processed frame 
    guint8* current_buffer;
    gint wh;
    /* Still need to setup encoding environment? */
    if (extrabuffer==NULL) {
        init_environment();
    }
    /* Our buffer to use */
    fast_memcpy(f_buffer_bak[0],f_buffer[0],width*height);
    fast_memcpy(f_buffer_bak[1],f_buffer[1],width*height/4);
    fast_memcpy(f_buffer_bak[2],f_buffer[2],width*height/4);
    current_buffer = f_buffer_bak[0];
    current_buffer_planar[0] = f_buffer_bak[0];
    current_buffer_planar[1] = f_buffer_bak[1];
    current_buffer_planar[2] = f_buffer_bak[2];
    /* Paused? */
    while (encoding_paused) {
        sleep(1);
    }
    /* Stop? */
    if (stopped) {
        stopped = FALSE;
        #ifdef STANDALONE
        printf("\tstopped.\n");
        #endif
        todo_frames = totalframes;
        quit = TRUE;
    }

    /* Adjust outputted .avi SCR value, based on video frames. Drip\avifile generated avi's
       are static framerate , so SCR += SCR + SCRframe */
    SCRdrip += SCRdrip + SCRframe;

    /* --- Plugin Filters, type PRE VIDEO --- */
    for (gint i=0;i<filters_index;i++) {
        if (filters[i].active==TRUE & filters[i].module_type==VIDEO && filters[i].module_phase==PRE) {
            current_buffer = (*filters[i].module_apply_function)(current_buffer,width*height,(gulong)SCRdrip,PTSframe);
            current_buffer_planar[0] = current_buffer;
            current_buffer_planar[1] = current_buffer+width*height;
            current_buffer_planar[2] = current_buffer+width*height*5/4;
        }
    }

    /* Deinterlace PAL */
    if (deinterlace==TRUE) {
        deinterlace_apply(deinterlace_buffer,current_buffer_planar[0],width,height);
        current_buffer = deinterlace_buffer;
        current_buffer_planar[0] = deinterlace_buffer;
    }
    /* Deinterlace NTSC (reverse 3:2 pulldown) */
    /* !!!! TODO : finish, this is incomplete code !!! */
    /* !!!!        Needs audio encoding to work 1st !! */
    if (pulldown==TRUE) {
        static guchar* tmpbuffer = inverse_pulldown(current_buffer_planar[0],width,height,totalframes);
        if (tmpbuffer!=current_buffer_planar[0] && tmpbuffer) {
            current_buffer = tmpbuffer;
            current_buffer_planar[0] = tmpbuffer;
            current_buffer_planar[1] = tmpbuffer+width*height;
            current_buffer_planar[2] = tmpbuffer+width*height*5/4;
        }
    }
    /* Subpicture */
    if (subpicture>-1) {
        spu_overlay(current_buffer_planar,width,height,clipbottom);
    }
    /* Scaling */
    if (scale==TRUE) {
        clip_frame_i420(current_buffer_planar, clippedbuffer,clipleft,cliptop,width-(clipleft+clipright), height-(cliptop+clipbottom),width,height);
        scale_frame_i420(clippedbuffer,(guchar*)scaledbuffer[0],width-(clipleft+clipright), height-(cliptop+clipbottom), owidth, oheight, clipleft, clipright, cliptop, clipbottom);
        current_buffer = scaledbuffer[0];
        current_buffer_planar[0] = scaledbuffer[0];
        current_buffer_planar[1] = scaledbuffer[1];
        current_buffer_planar[2] = scaledbuffer[2];
    }

    /* --- Plugin Filters, type NORMAL VIDEO --- */
    for (gint i=0;i<filters_index;i++) {
        if (filters[i].active==TRUE & filters[i].module_type==VIDEO && filters[i].module_phase==NORMAL) {
            current_buffer = (*filters[i].module_apply_function)(current_buffer,owidth*oheight,(gulong)SCRdrip,PTSframe);
            current_buffer_planar[0] = current_buffer;
            current_buffer_planar[1] = current_buffer+owidth*oheight;
            current_buffer_planar[2] = current_buffer+owidth*oheight*5/4;
        }
    }

    /* --- Plugin Filters, type POST VIDEO --- */
    for (gint i=0;i<filters_index;i++) {
        if (filters[i].active==TRUE && filters[i].module_type==VIDEO && filters[i].module_phase==POST) {
            current_buffer = (*filters[i].module_apply_function)(current_buffer,owidth*oheight,(gulong)SCRdrip,PTSframe);
            current_buffer_planar[0] = current_buffer;
            current_buffer_planar[1] = current_buffer+owidth*oheight;
            current_buffer_planar[2] = current_buffer+owidth*oheight*5/4;
        }
    }

    /* -- HANDLE PROCESSED FRAME -- */
    if (current_buffer!=scaledbuffer[0]) {
        fast_memcpy(scaledbuffer[0],current_buffer_planar[0],owidth*oheight);
        fast_memcpy(scaledbuffer[1],current_buffer_planar[1],owidth*oheight/4);
        fast_memcpy(scaledbuffer[2],current_buffer_planar[2],owidth*oheight/4);
    }
    /* Store current frame to frame buffer */
    encode_frame(stream,im,FALSE,FALSE);
    /* Add audio */
    output_audio(totalframes);

    /* -- FEEDBACK -- */
    /* Notify GUI with progress every ?100? frames */
    if ((totalframes_countdown--)==0) {
        totalframes_countdown = 24;
        #ifndef STANDALONE
        gui_progress(totalframes,todo_frames);//,j,numfiles);
        #else
        fprintf(stderr,"\rEncoding frame %i of %i",totalframes,todo_frames);
        #endif
    }
    #ifndef STANDALONE
    /* Writeout current frame, at every 500 frames minimal */
    if ( (writeout_frames_countdown--) == 0 ) {
        /* Next writeout */
        writeout_frames_countdown = writeout_frames_countdown_amount;
        if (writeout_frames_countdown_amount<1000) {
            writeout_frames_countdown_amount+=25;
        } else {
            writeout_frames_countdown_amount = 499;
        }
        /* Convert current frame to RGB */
        I420toRGB24((guint8*)ToRGBbuffer,(guchar*)scaledbuffer[0],bminfo_in.biWidth,labs(bminfo_in.biHeight));
        /* Writeout frame */
        gui_frame(ToRGBbuffer,totalframes,bminfo_in.biWidth,labs(bminfo_in.biHeight));
    }
    #endif

    /* Reached endframe? */
    if (totalframes > (todo_frames-1)) {
        quit = TRUE;
    }
    return;
}


/*
 * Main encoding function
 */
gpointer encode(gpointer data) {
    arg_parmS *arg_parm = (arg_parmS*)data;
    gint argc = arg_parm->argc;
    gchar **argv = arg_parm->argv;
    GString *eta = g_string_new("");
    time_t Time;// = time(NULL);
    gulong seconds;
    gulong minutes;
    gulong hours;
    g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Starting encoding process");
    /* Renice */
    if ((arg_parm->nice<=20) && (arg_parm->nice>=20)) {
        result = nice(arg_parm->nice);
        if (result!=0) {
            g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"Renice failed, used an invalid value");
        }
    }
    /* Check params */
    if(argc < 4) {
        display_help(argv[0]);
    }
    /* Fill audio structure for both channels */
    for (gint anum=0;anum<AMAX;anum++) {
        au[anum].in_stream = -1;
        au[anum].codec = 0x55;   // Default to MP3 audio 
        au[anum].mp3bitrate = -1;
        au[anum].achan = 0;
        au[anum].rate = 0;
        au[anum].samples = 0;
        au[anum].outstream = NULL;
        au[anum].volume_adjustment = 1; 
    }
    /* Set audio channel 0 to default */
    au[0].in_stream = 0;
    /* Fill codec structure */
    opendivx_params.rc_period=0;
    opendivx_params.rc_reaction_period=0;
    opendivx_params.rc_reaction_ratio=0;
    opendivx_params.max_key_interval=0;
    opendivx_params.obmc=0;
    opendivx_params.quality=5;
    opendivx_params.max_quantizer=12;
    opendivx_params.min_quantizer=2;
    /* Parse parameters */
    moreopts = TRUE;
    argbase = 0;
    while(moreopts) {
        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Handling parameter %s",argv[argbase+1]);
        moreopts=FALSE;
        if(strcmp(argv[argbase+1],"-cl")==0) {
            clipleft=atoi(argv[argbase+2]);
            argbase+=2;
            moreopts=TRUE;
        }
        if(strcmp(argv[argbase+1],"-cr")==0) {
            clipright=atoi(argv[argbase+2]);
            argbase+=2;
            moreopts=TRUE;
        }
        if(strcmp(argv[argbase+1],"-b")==0) {
            divxbitrate=atoi(argv[argbase+2]);
printf("bitrate = %i\n",divxbitrate);
            argbase+=2;
            moreopts=TRUE;
        }
        if(strcmp(argv[argbase+1],"-f")==0) {
            video_codec=fccDIV4;
            argbase++;
            moreopts=TRUE;
        }
        #if (AVIFILE_MAJOR_VERSION == 0) && (AVIFILE_MINOR_VERSION==6) 
        if(strcmp(argv[argbase+1],"-d")==0) {
            video_codec=fccDIVX;//mmioFOURCC('d', 'i', 'v', '3');//1482049860;
            argbase++;
            moreopts=TRUE;
        }
        #endif
        if(strcmp(argv[argbase+1],"-dq")==0) {
            opendivx_params.quality=atoi(argv[argbase+2]);
            argbase++;
            moreopts=TRUE;
        }
        if(strcmp(argv[argbase+1],"-dminq")==0) {
            opendivx_params.min_quantizer=atoi(argv[argbase+2]);
            argbase++;
            moreopts=TRUE;
        }
        if(strcmp(argv[argbase+1],"-dmaxq")==0) {
            opendivx_params.max_quantizer=atoi(argv[argbase+2]);
            argbase++;
            moreopts=TRUE;
        }
        if(strcmp(argv[argbase+1],"-a")==0) {
            au[0].mp3bitrate=atoi(argv[argbase+2]);
            argbase+=2;
            moreopts=TRUE;
        }
        if(strcmp(argv[argbase+1],"-a2")==0) {
            au[1].mp3bitrate=atoi(argv[argbase+2]);
            argbase+=2;
            moreopts=TRUE;
        }
        if(strcmp(argv[argbase+1],"-av")==0) {
            au[0].volume_adjustment=atoi(argv[argbase+2]);
            argbase+=2;
            moreopts=TRUE;
        }
        if(strcmp(argv[argbase+1],"-av2")==0) {
            au[1].volume_adjustment=atoi(argv[argbase+2]);
            argbase+=2;
            moreopts=TRUE;
        }
        if(strcmp(argv[argbase+1],"-r")==0) {
            aspect=atof(argv[argbase+2]);
            argbase+=2;
            moreopts=TRUE;
        }
        if(strcmp(argv[argbase+1],"-e")==0) {
            endframe=atoi(argv[argbase+2]);
            argbase+=2;
            moreopts=TRUE;
        }
        if(strcmp(argv[argbase+1],"-c")==0) {
            video_codec=atoi(argv[argbase+2]);
            argbase+=2;
            moreopts=TRUE;
        }
        if(strcmp(argv[argbase+1],"-au")==0) {
            au[0].codec=atoi(argv[argbase+2]);
            argbase+=2;
            moreopts=TRUE;
        }
        if (AMAX>1)
            if(strcmp(argv[argbase+1],"-au2")==0) {
            au[1].codec=atoi(argv[argbase+2]);
            argbase+=2;
            moreopts=TRUE;
        }
        if(strcmp(argv[argbase+1],"-ct")==0) {
            cliptop=atoi(argv[argbase+2]);
            argbase+=2;
            moreopts=true;
        }
        if(strcmp(argv[argbase+1],"-cb")==0) {
            clipbottom=atoi(argv[argbase+2]);
            argbase+=2;
            moreopts=true;
        }
        if(strcmp(argv[argbase+1],"-ip")==0) {
            deinterlace = TRUE;
            argbase+=1;
            moreopts=true;
            g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Activated PAL deinterlace filter");
        }
        //        if(strcmp(argv[argbase+1],"-in")==0) {
        //            pulldown = TRUE;
        //            argbase+=1;
        //            moreopts=true;
        //            g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Activated NTSC reverse 3:2 pulldown. NOT YET READY!");
        //        }
        if(strcmp(argv[argbase+1],"-ac")==0) {
            autoclip = TRUE;
            argbase+=1;
            moreopts=true;
        }
        if(strcmp(argv[argbase+1],"-w")==0) {
            owidth=atoi(argv[argbase+2]);
            scale = TRUE;
            argbase+=2;
            moreopts=true;
        }
        if(strcmp(argv[argbase+1],"-h")==0) {
            oheight=atoi(argv[argbase+2]);
            scale = TRUE;
            argbase+=2;
            moreopts=true;
        }
        if(strcmp(argv[argbase+1],"-as")==0) {
            au[0].in_stream=atoi(argv[argbase+2]);
            argbase+=2;
            moreopts=true;
        }
        if (AMAX>1) {
            if(strcmp(argv[argbase+1],"-as2")==0) {
                au[1].in_stream=atoi(argv[argbase+2]);
                argbase+=2;
                moreopts=true;
            }
        }
        if(strcmp(argv[argbase+1],"-vs")==0) {
            vistream=atoi(argv[argbase+2]);
            argbase+=2;
            moreopts=true;
        }
        if(strcmp(argv[argbase+1],"-s")==0) {
            subpicture=atoi(argv[argbase+2]);
            clutfile= argv[argbase+3];
            argbase+=3;
            moreopts=true;
        }
        if(strcmp(argv[argbase+1],"-l")==0) {
            limit=1;
            max_size=atoi(argv[argbase+2]);
            argbase+=2;
            moreopts=true;
        }

        if(strcmp(argv[argbase+1],"-A")==0) {
            argbase+=1;
            moreopts=true;
        }
        if(strcmp(argv[argbase+1],"-F")==0) {
            argbase+=1;
            moreopts=true;
        }
    }
    /* Chunk size */
    if (max_size<50000) {
        /* Stop polling loops */
        encoding = TRUE;
        usleep(15000);
        #ifndef STANDALONE
        gui_ready();
        #endif
        encoding = FALSE;
        g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"Max size too small, use at least 50000kb");
        return 0;
    } else {
        /* Adjust max size so max_size_new + header =~ max_size */
        max_size=max_size-50000;
    }
    /* Input files */
    numfiles=argc-argbase-2;
    for(gint i=argbase+1; i<argc-1; i++) {
        inputfiles[i-argbase-1] = g_string_new(argv[i]);
        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Input %d out %s",i-argbase-1,inputfiles[i-argbase-1]->str);
    }
    if ((inputfiles[0] == NULL) || (inputfiles[0]->str == NULL)) {
        g_log(DRIP_LD,G_LOG_LEVEL_ERROR,"Input or Output filename missing in parameters");
    }
    outputfile=argv[argc-1];
    if (outputfile == NULL) {
        g_log(DRIP_LD,G_LOG_LEVEL_ERROR,"Output filename missing in parameters");
    }
    /* Todo frames */
    todo_frames = endframe;
    /* Various varis' */
    output = g_string_new("");
    framefile = g_string_new("");
    g_string_sprintf(framefile,"%s/.drip/frame.raw",getenv("HOME"));
    frame_buffer_index = 0;
    frame_buffer_base = 0;
    /* Init filters */
    autoclipping();
    /* Notify user begin status, and fetch start time */
    #ifndef STANDALONE
    gui_progress(0,todo_frames);//,j,numfiles);
    #else
    fprintf(stderr,"\rEncoding frame %i of %i",0,todo_frames);
    #endif 
    /* Inint time */
    gettimeofday(&tv,&tz); /* Get start time for FPS */
    startsec=tv.tv_sec;
    startusec=tv.tv_usec;
    totalframes = 0;

    /* 
     * Encoding starts
     */

    /* Open mpeg stream (video+audio) */
    g_log(DRIP_LD,G_LOG_LEVEL_MESSAGE,"Opening %s",inputfiles[0]->str);
    file = drip_mpeg_open(inputfiles[0]->str,process_frame);
    if (!file) {
        /* Error, not able to open stream properly */
        g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"Unable to open %s, possibly not a plain MPEG2 or valid IFO file",inputfiles[0]->str);
        encoding = TRUE;
        #ifndef STANDALONE
        gui_ready();
        #endif
        encoding = FALSE;
        return 0;
    }
    encoding = TRUE;
    g_log(DRIP_LD,G_LOG_LEVEL_MESSAGE,"Opening input AV stream from %s",inputfiles[0]->str);
    if (!file) {
        /* Error, not able to open input stream properly */
        g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"Unable to open %s, possibly not a plain MPEG2 or valid IFO file",inputfiles[0]->str);
        encoding = TRUE;
        #ifndef STANDALONE
        gui_ready();
        #endif
        /* Close streams */
        stream->Stop();
        delete avifile;
        drip_mpeg_close();
        /* Encoding done */
        encoding = FALSE;
    }
    /* Reset total frames counter */
    totalframes_countdown = 25;
    /* -- Start demuxer (decoding + filtering + encoding process) -- */
    quit = FALSE;
    drip_mpeg_start_demuxing(au[0].in_stream); //TODO 2nd audio channel
    /* Stop & close encoding */
    #ifndef STANDALONE
    drip_lock("Destructing video encoding codec");
    #endif
    stream->Stop();
    #ifndef STANDALONE
    drip_unlock("Destructing video encoding codec");
    drip_lock("Destructing encoder object");
    #endif
    delete avifile;
    #ifndef STANDALONE
    drip_unlock("Destructing encoder object");
    drip_lock("Destructing mpeg decoder");
    #endif
    drip_mpeg_close();
    #ifndef STANDALONE
    drip_unlock("Destructing mpeg decoder");
    #endif
    /* Display encoding info */
    gettimeofday(&tv,&tz);
    fps=totalframes/((tv.tv_sec+tv.tv_usec/1000000.0)-(startsec+startusec/1000000.0));
    seconds = (gulong)(tv.tv_sec-startsec);
    hours = (guint)(seconds/3600);
    seconds -= 3600*hours;
    minutes = (guint)(seconds/60);
    seconds -= 60*minutes;
    eta = g_string_new("");
    if (hours>0) {
        g_string_sprintf(eta,"%li hours %li minutes",hours,minutes);
    } else {
        g_string_sprintf(eta,"%li minutes %li seconds",minutes,seconds);
    }
    #ifdef STANDALONE
    fprintf(stderr,"\n");
    #endif
    g_log(DRIP_LD,G_LOG_LEVEL_MESSAGE,"Finished, %li frames in %s at %.2g fps",totalframes,eta->str,fps);
    //g_string_free(output,TRUE);
    /* Notify GUI encoding has finished */
    #ifndef STANDALONE
    gui_ready();
    #endif
    /* Clean & Exit */
    //free(framebuffer);framebuffer = NULL;
    //free(extrabuffer);
    extrabuffer = NULL;
    //delete(rowptr);
    //free(clippedbuffer[0]);
    //free(clippedbuffer[1]);
    //free(clippedbuffer[2]);
    //delete(scaledbuffer);
    //free(_RGBbuffer);
    //delete(im);
    if (pulldown==TRUE) pulldown_cleanup();
    //delete(progress_frame_im);
    //g_string_free(eta,TRUE);
    /* Deallocate frame buffer */
    //for (gint i=0;i<frame_buffer_index_MAX;i++) {
    //    free(frame_buffer[i].planes[0]);
    //}
    //frame_raw = NULL;
    frame_buffer_index = -1;
    frame_buffer_base = 0;
    writeout_frames_countdown_amount = 10;
    writeout_frames_countdown = 10;
    /* Destruct plugins (NOT unload them) */
    close_plugins();
    /* Encoding done */
    encoding = FALSE;
    audio_output_codec_setup = FALSE;
    first_frame = TRUE;
    #ifdef STANDALONE
    return NULL;
    #else
    pthread_exit(0);
    #endif
}

