// Jarl van Katwijk - 2001
// licensed under the GNU Public License (GPL) v2

#include "main.hh"
#include <glib.h>
#include "encoder.hh"
#include <stream.h>
#include <iostream>
#include <vector>
#include <dirent.h>
#include <dlfcn.h>
#include <videoencoder.h>
#include <avifile.h>
#include <aviplay.h>
#include <version.h>

#include <ostream.h>
#include <fstream>
#include <strstream>
#include <cstdlib>
#include <cstdio>
#include <string.h>
#include <dirent.h>
#include <dlfcn.h>
#include <string>

#if (AVIFILE_MAJOR_VERSION == 0) && (AVIFILE_MINOR_VERSION==6) 
   #include <default.h>
   #include <fourcc.h>
   #include <creators.h>
#else
   #include <aviutil.h>
#endif

using namespace std;

extern "C" {
    codecS **list;

    void encoder_start(const char *parameters,glong nice) {
        /* start encoding */
        gint count = 0;
        gchar **param;
        arg_parmS arg_parm;
        gpointer temp;

        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Got activated");
        /* already encoding? */
        if (encoding) {
            g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"Already encoding.");
            return;
        }

        /* split parameters */
        param = g_strsplit(parameters," ",0);
        while (param[count]!=NULL) {
            count++;
        }
        arg_parm.argc = count;
        arg_parm.argv = param;
        arg_parm.nice = nice;

        /* do the encoding */
        temp = &arg_parm; //keep compiler happy about next line :
        encoderthread_id = pthread_create(&encoderthread,NULL,encode,temp);

        /* wait for parsing to be done and real encoding to start */
        while (!encoding) {
            usleep(100000);
        }

        /* free & exit */
        return;
    }

    void encoder_quit(gboolean lock) {
        gchar *domain;

        if (lock) domain = DRIP_LD;
        else domain = DRIP_CB_LD;
        /* TODO: cleanup before exit */
        if (encoding) {
            pthread_cancel(encoderthread);
            encoding = FALSE;
        }
        /* remove ior file */
        unlink("dripencoder.ior");
        g_log(domain,G_LOG_LEVEL_MESSAGE,"dripencoder exiting.");
        return;
    }
    void encoder_stop(gboolean lock) {
        gchar *domain;

        if (lock) domain = DRIP_LD;
        else domain = DRIP_CB_LD;

        if (encoding) {
            stopped = TRUE;
            encoding = FALSE;
            g_log(domain,G_LOG_LEVEL_DEBUG,"Stopped encoding");
        } else {
            g_log(domain,G_LOG_LEVEL_DEBUG,"Not encoding, couldn't stop");
        }
        return;
    }

    void encoder_pause(gboolean lock) {
        gchar *domain;

        if (lock) domain = DRIP_LD;
        else domain = DRIP_CB_LD;

        if (encoding) {
            encoding_paused = TRUE;
            g_log(domain,G_LOG_LEVEL_MESSAGE,"Paused encoding ...");
        } else {
            g_log(domain,G_LOG_LEVEL_DEBUG,"Not encoding, couldn't pause");
        }
        return;
    }

    void encoder_resume(gboolean lock) {
        gchar *domain;

        if (lock) domain = DRIP_LD;
        else domain = DRIP_CB_LD;

        if (encoding && encoding_paused) {
            encoding_paused = FALSE;
            g_log(domain,G_LOG_LEVEL_MESSAGE,"... Resumed proces");
        } else {
            g_log(domain,G_LOG_LEVEL_DEBUG,"Not paused, couldn't resume");
        }
        return;
    }

    codecS** encoder_codecs() {
        avm::vector<CodecInfo>::iterator it;
        gint codec_nr = 0,i;
        gint codec=0x55;
        GString *temp = g_string_new("");

        /* Set an attribute so codecs are loaded */
#if (AVIFILE_MAJOR_VERSION == 0) && (AVIFILE_MINOR_VERSION==6) 
        const CodecInfo* codecInfo=CodecInfo::match(codec);
//        Creators::SetCodecAttr(*codecInfo, "Crispness", 100);
#else
//        IVideoEncoder::SetExtendedAttr(codec, "Crispness", 100);
#endif
        /* Scan for amount of plugins */
        for (it = video_codecs.begin(); it != video_codecs.end(); it++) {
             if (it->direction & CodecInfo::Encode) {
                 avm::vector<fourcc_t>::const_iterator iv;
                 for (iv = it->fourcc_array.begin(); iv != it->fourcc_array.end(); iv++) {
                     codec_nr++;
                 }
             }
        }
        for (it = audio_codecs.begin(); it != audio_codecs.end(); it++) {
              avm::vector<fourcc_t>::const_iterator ia;
              for (ia = it->fourcc_array.begin(); ia != it->fourcc_array.end(); ia++) {
                  codec_nr++;
              }
        }
        /* Build codec list */
        list = (codecS**)malloc(sizeof(codecS*) * codec_nr);
        list[0] = (codecS*)malloc(sizeof(codecS));
        /* Alloc list */
        for (i=1;i<codec_nr;i++) {
            list[i] = (codecS*)malloc(sizeof(codecS));
        }

        codec_nr = 0;
        for (it = video_codecs.begin(); it != video_codecs.end(); it++) {
            if (it->direction & CodecInfo::Encode) {
                avm::vector<fourcc_t>::const_iterator iv;
                for (iv = it->fourcc_array.begin(); iv != it->fourcc_array.end(); iv++) {
                    list[codec_nr]->ctype = g_string_new("    ");
                    list[codec_nr]->ctype->str[0] = ((const char*)&(*iv))[0];
                    list[codec_nr]->ctype->str[1] = ((const char*)&(*iv))[1];
                    list[codec_nr]->ctype->str[2] = ((const char*)&(*iv))[2];
                    list[codec_nr]->ctype->str[3] = ((const char*)&(*iv))[3];
                    list[codec_nr]->ccodec = *iv;
                    list[codec_nr]->cmodule = g_string_new(it->modulename.c_str());
                    list[codec_nr]->cav = (enum codecT)0;
                    codec_nr++;
                }
            }
        }
        for (it = audio_codecs.begin(); it != audio_codecs.end(); it++) {
            if (it->direction & CodecInfo::Encode) {
                avm::vector<fourcc_t>::const_iterator ia;
                for (ia = it->fourcc_array.begin(); ia != it->fourcc_array.end(); ia++) {
                    list[codec_nr]->ctype = g_string_new("    ");
                    list[codec_nr]->ctype->str[0] = ((const char*)&(*ia))[0];
                    list[codec_nr]->ctype->str[1] = ((const char*)&(*ia))[1];
                    list[codec_nr]->ctype->str[2] = ((const char*)&(*ia))[2];
                    list[codec_nr]->ctype->str[3] = ((const char*)&(*ia))[3];
                    list[codec_nr]->ccodec = *ia;
                    list[codec_nr]->cmodule = g_string_new(it->modulename.c_str());
                    list[codec_nr]->cav = (enum codecT)1;
                    codec_nr++;
                }
            }
        }

        list[0]->length = codec_nr;
        g_string_free(temp,TRUE);
        return list;
    }

} // extern "C"
