
#include "../src/drip.h"
#include "plugin-loader.hh"
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <string.h>


filter filters[MAX_FILTERS]; // list of filters
gint filters_index = 0; // index to filters array

/* Save settings of a plugin */
gboolean save_plugin_settings(gint plugin) {
    GIOChannel *configfile_channel = NULL;
    guint bytessaved;
    gint saveerror;
    gint configfile_fd;
    gint len;
    gboolean result = TRUE;
    /* Fetch values */
    gfloat values[5];
    gboolean active = filters[plugin].active;
    gchar* name = filters[plugin].module_name->str;
    GString *HOME = g_string_new(getenv("HOME"));
    GString *configfile = g_string_new("");
    /* Create (and remove old) configuration file */
    (*filters[plugin].module_values_function_get)(values);
    g_string_sprintf(configfile,"%s/.drip/plugins/%s.cfg",HOME->str,name);
    unlink(configfile->str);                                                 // delete old
    configfile_fd = open(configfile->str,O_WRONLY|O_CREAT,S_IREAD|S_IWRITE); // create config file
    if (configfile_fd>-1) {
        configfile_channel = g_io_channel_unix_new(configfile_fd);
        len = strlen(name);
        saveerror = g_io_channel_write(configfile_channel,(gchar*)&len,sizeof(gint),&bytessaved);
        saveerror = g_io_channel_write(configfile_channel,name,len,&bytessaved);
        saveerror = g_io_channel_write(configfile_channel,(gchar*)&active,sizeof(gboolean),&bytessaved);
        saveerror = g_io_channel_write(configfile_channel,(gchar*)&values[0],sizeof(gfloat),&bytessaved);
        saveerror = g_io_channel_write(configfile_channel,(gchar*)&values[1],sizeof(gfloat),&bytessaved);
        saveerror = g_io_channel_write(configfile_channel,(gchar*)&values[2],sizeof(gfloat),&bytessaved);
        saveerror = g_io_channel_write(configfile_channel,(gchar*)&values[3],sizeof(gfloat),&bytessaved);
        saveerror = g_io_channel_write(configfile_channel,(gchar*)&values[4],sizeof(gfloat),&bytessaved);
        g_io_channel_close(configfile_channel);
        close(configfile_fd);
    } else {
        result = FALSE;
    }
    /* Clean & exit */
    g_string_free(configfile,TRUE);
    g_string_free(HOME,TRUE);
    return result;
}

/* Load settings of plugins */
gboolean load_plugin_settings(gint plugin) {
    GIOChannel *configfile_channel = NULL;
    guint bytesloaded;
    gint loaderror;
    gint configfile_fd;
    gboolean result = TRUE;
    gfloat values[5];
    gboolean active;
    gchar* name = filters[plugin].module_name->str;
    gint len;
    gpointer buf = malloc(1024);
    GString *HOME = g_string_new(getenv("HOME"));
    GString *configfile = g_string_new("");
    memset(buf,0,1024);
    /* Open configuration file */
    g_string_sprintf(configfile,"%s/.drip/plugins/%s.cfg",HOME->str,name);
    configfile_fd = open(configfile->str,O_RDONLY); // create config file
    if (configfile_fd>-1) {
        configfile_channel = g_io_channel_unix_new(configfile_fd);
        loaderror = g_io_channel_read(configfile_channel,(gchar*)&len,sizeof(gint),&bytesloaded);
        loaderror = g_io_channel_read(configfile_channel,(gchar*)buf,len,&bytesloaded);
        loaderror = g_io_channel_read(configfile_channel,(gchar*)&active,sizeof(gboolean),&bytesloaded);
        loaderror = g_io_channel_read(configfile_channel,(gchar*)&values[0],sizeof(gfloat),&bytesloaded);
        loaderror = g_io_channel_read(configfile_channel,(gchar*)&values[1],sizeof(gfloat),&bytesloaded);
        loaderror = g_io_channel_read(configfile_channel,(gchar*)&values[2],sizeof(gfloat),&bytesloaded);
        loaderror = g_io_channel_read(configfile_channel,(gchar*)&values[3],sizeof(gfloat),&bytesloaded);
        loaderror = g_io_channel_read(configfile_channel,(gchar*)&values[4],sizeof(gfloat),&bytesloaded);
        g_io_channel_close(configfile_channel);
        close(configfile_fd);
    } else {
        active = FALSE;
        values[0] = 0;
        values[1] = 0;
        values[2] = 0;
        values[3] = 0;
        values[4] = 0;
        result = FALSE;
    }
    /* Set values into plugin */
    filters[plugin].active = active;
    (*filters[plugin].module_values_function_set)(values[0],values[1],values[2],values[3],values[4]);
    /* Clean & exit */
    g_string_free(configfile,TRUE);
    g_string_free(HOME,TRUE);
    free(buf);
    return TRUE;
}

/* Load a symbol out a module */
gboolean load_symbol(GModule *module,gchar *name,gpointer *symbol) {
    static gboolean result;
    result = g_module_symbol(module,name,symbol);
    if (result==FALSE) {
        g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"Plugin: loading failed, cant find symbol %s",name);
        return FALSE;
    } else 
        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Plugin: found symbol %s",name);
    return TRUE;
}

/* Loads a plugin, checks it and updates list-of-loaded plugins */
gboolean load_plugin(gchar* plugin,gchar* directory) {
    GString* symbol_name = g_string_new("");
    gboolean result;
    gpointer *symbol;
    
    /* Construct main symbol name */
    filters[filters_index].module_name = g_string_new("");
    g_string_sprintf(filters[filters_index].module_name,"drip_%s",plugin);
    /* Load & Scan module */
    filters[filters_index].module_filename = g_module_build_path((const gchar*)directory,(const gchar*)filters[filters_index].module_name->str);
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Plugin: loading module %s",filters[filters_index].module_filename);
    filters[filters_index].module = g_module_open((const gchar*)filters[filters_index].module_filename,G_MODULE_BIND_MASK);

    /* Module loaded? */
    if (filters[filters_index].module==NULL) {
        g_log(DRIP_LD,G_LOG_LEVEL_WARNING,"Plugin: loading module %s failed",filters[filters_index].module_filename);
        return FALSE;
    }

    /* -- Load symbols -- */
    /* Init Symbol in module? */
    g_string_sprintf(symbol_name,"%s_init",plugin);
    result = load_symbol(filters[filters_index].module,symbol_name->str,(gpointer*)(&filters[filters_index].module_init_function));
    if (result==FALSE) return FALSE;
    /* Apply Symbol in module? */
    g_string_sprintf(symbol_name,"%s_apply",plugin);
    result = load_symbol(filters[filters_index].module,symbol_name->str,(gpointer*)(&filters[filters_index].module_apply_function));
    if (result==FALSE) return FALSE;
    /* GUI Symbol in module? */
    g_string_sprintf(symbol_name,"%s_gui",plugin);
    result = load_symbol(filters[filters_index].module,symbol_name->str,(gpointer*)(&filters[filters_index].module_gui_function));
    if (result==FALSE) return FALSE;
    /* Apply Symbol in module? */
    g_string_sprintf(symbol_name,"%s_destruct",plugin);
    result = load_symbol(filters[filters_index].module,symbol_name->str,(gpointer*)(&filters[filters_index].module_destruct_function));
    if (result==FALSE) return FALSE;
    /* Type Symbol in module? */
    g_string_sprintf(symbol_name,"%s_type",plugin);
    result = load_symbol(filters[filters_index].module,symbol_name->str,(gpointer*)(&filters[filters_index].module_type_function));
    if (result==FALSE) return FALSE;
    /* Phase Symbol in module? */
    g_string_sprintf(symbol_name,"%s_phase",plugin);
    result = load_symbol(filters[filters_index].module,symbol_name->str,(gpointer*)(&filters[filters_index].module_phase_function));
    if (result==FALSE) return FALSE;
    /* Values Symbol in module? */
    g_string_sprintf(symbol_name,"%s_values",plugin);
    result = load_symbol(filters[filters_index].module,symbol_name->str,(gpointer*)(&filters[filters_index].module_values_function_get));   
    if (result==FALSE) return FALSE;
    /* Values_function Symbol in module? */
    g_string_sprintf(symbol_name,"%s_values_function",plugin);
    result = load_symbol(filters[filters_index].module,symbol_name->str,(gpointer*)(&filters[filters_index].module_values_function_set));
    if (result==FALSE) return FALSE;



    /* Fill in filter info */
    filters[filters_index].loaded = TRUE;
    filters[filters_index].active = FALSE; // TODO: Set to TRUE based on configuration\plugin dialogs window
    filters[filters_index].module_GUI = NULL;
    filters[filters_index].module_type = (*filters[filters_index].module_type_function)();
    filters[filters_index].module_phase = (*filters[filters_index].module_phase_function)();

    /* Update filters_index */
    filters_index++;

    /* Clean & Exit */
    g_string_free(symbol_name,TRUE);
    return result;
}


/* Close all plugins (save their config, and call module_destruct_function) */
#ifdef __cplusplus
extern "C" {
#endif
gboolean close_plugins(void) {
    gboolean result = TRUE;
    
    g_log(DRIP_CB_LD,G_LOG_LEVEL_DEBUG,"Plugin: closing plugins");
    /* Destruct plugins (NOT unload them) */
    for (gint i=0;i<filters_index;i++) {
        result = result & save_plugin_settings(i);
        (*filters[i].module_destruct_function)();
    }
 
    if (result==TRUE) {
        g_log(DRIP_CB_LD,G_LOG_LEVEL_DEBUG,"Closed %i plugins succesfully",filters_index);
    } else { 
        g_log(DRIP_CB_LD,G_LOG_LEVEL_DEBUG,"Closing plugins failed, read log for details");
    }
    /* --- */
    
    return result;
}   
#ifdef __cplusplus
}   
#endif


/* Initialize plugin environment, and loads them */
#ifdef __cplusplus
extern "C" {
#endif
gboolean init_plugins(void) {
    gboolean result;
    /* Can we load plugins at all? */
    result = g_module_supported();
    if (result == FALSE) {
        return FALSE;
    }
    /* --- Load the filters
       TODO: scan lib drip for plugins, dont use hardcoded table */
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Plugin: loading plugins");
    result = load_plugin("testfilter","/usr/local/lib");
    result = result & load_plugin("noisefilter","/usr/local/lib");
    result = result & load_plugin("volumefilter","/usr/local/lib");
    /* Load settings */
    for (gint i=0;i<filters_index;i++) {
        result = result & load_plugin_settings(i);
    }
    /* --- */
    if (result==TRUE) {
        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Loaded %i plugins succesfully",filters_index);
    } else {
        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Loading plugins failed, read log for details");
    }
    /* --- */
    return result;
}
#ifdef __cplusplus
}
#endif

