#include "../config.h"
#include "../src/drip.h"
#include <gmodule.h>
#include "../encoder/plugin-loader.hh"
// Custom includes:
#include "../encoder/fast_memcpy.hh"

static gboolean init_done = FALSE;                 // Plugin initialised flag
static gint width,height;                          // Input Width\Height sizes
static gint xoff,yoff,xlen,ylen;                   // Offset and lenght values
static gfloat framerate;                           // Framerate (25.00, 29.97, etc)
static gfloat value1,value2,value3,value4,value5;  // Plugin defined values
static guint8 *src_planes[3];                      // Input data planes (I420 colour space)
static guint8 *dst_planes[3];                      // Output data planes
// Custom vari's
static guint8* backup[3];
static gchar noise_T[800];
static glong wh,wh4;
gfloat noise_T1164[256];
gfloat noise_T2018[256];
gfloat noise_T0813[256];
gfloat noise_T1596[256];
gfloat noise_T0391[256];


extern "C" {

/* Limit colour to valid range */
inline guint8 normalize_colour(gfloat colour) {
    static guint colour_norm;

    if (colour<0) {
        colour_norm = 0;
    } else if (colour>255) {
        colour_norm = 255;
    } else {
        colour_norm = (guint8)(colour);
    }
    return colour_norm;
}

/* Return type (SPU,AUDIO,VIDEO) of filter */
G_MODULE_EXPORT module_type_e noisefilter_type(void) {
    return VIDEO;
}


/* Return phase (PRE,NORMAL,POST) of filter */
G_MODULE_EXPORT module_phase_e noisefilter_phase(void) {
    return POST;
}

/* Return values of plugin (value1,value2,... value5) */
/* NOTE: calling function must deallocate the returned gfloat[5] */
G_MODULE_EXPORT void noisefilter_values(gfloat* values) {
    values[0] = value1;
    values[2] = value2;
    values[3] = value3;
    values[4] = value4;
    values[5] = value5;
    return;
}

/* Set plugins values */
G_MODULE_EXPORT void noisefilter_values_function(gfloat v1,gfloat v2,gfloat v3,gfloat v4,gfloat v5) {
    value1 = v1;
    value2 = v2;
    value3 = v3;
    value4 = v4;
    value5 = v5;
    return;
}

/* Function which implements a GtkPlug to let user configure this plugin */
G_MODULE_EXPORT GtkWidget* noisefilter_gui(void) {
    static GtkWidget *gui = NULL;
    g_log(DRIP_CB_LD,G_LOG_LEVEL_DEBUG,"Plugin: noisefilter gui");

    /* Already constructed gui? */
    if (gui==NULL) {
        gui = gtk_label_new("\tnoisefilter\nthis filter isn't finished yet, dont use it");
    }

    /* Clean & Exit */
    return gui;
}


/* XXX: no chroma interpolating is done */
G_MODULE_EXPORT guint8* noisefilter_apply(guint8* src, glong size, gulong SCR, gulong PTS) {
//    #ifdef DEBUG
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Plugin: noisefilter apply");
//    #endif
    static gint offset = 0;
    static gint offset4 = 0;
    static gint offset3 = 0;
    static gfloat offset4_f;
    static guint8 Ys,Yo,CRs,CRo,CBs,CBo;
    static gfloat Bfs,Gfs,Rfs,Bfo,Gfo,Rfo;
    static guint8 Bs,Gs,Rs,Bo,Go,Ro;
    static guint diff;
 
    offset = 0;
    offset3 = 0;
    /* Check static framerate */
    if (PTS != SCR) {
        // unstatic framerate.. 
    }
    /* Loop */
    for (gint y=0;y<height;y++) {
        offset4_f = (y>>1) * (width>>1);
        for (gint x=0;x<width;x++) {
             offset4 = (gint)offset4_f;
             Ys = src_planes[0][offset];
             CRs = src_planes[1][offset4];
             CBs = src_planes[2][offset4];

             Bfs = noise_T1164[Ys] + noise_T2018[CBs];
             Gfs = noise_T1164[Ys] - noise_T0813[CRs] - noise_T0391[CBs];
             Rfs = noise_T1164[Ys] + noise_T1596[CRs];

             Bs = normalize_colour(Bfs);
             Gs = normalize_colour(Gfs);
             Rs = normalize_colour(Rfs);


             Yo = backup[0][offset];
             CRo = backup[1][offset4];
             CBo = backup[2][offset4];

             Bfo = noise_T1164[Yo] + noise_T2018[CBo];
             Gfo = noise_T1164[Yo] - noise_T0813[CRo] - noise_T0391[CBo];
             Rfo = noise_T1164[Yo] + noise_T1596[CRo];

             Bo = normalize_colour(Bfo);
             Go = normalize_colour(Gfo);
             Ro = normalize_colour(Rfo);


             diff = abs(Rs-Ro);
             if (diff < value1) {
                 if (diff > (value2)) {
                     dst_planes[0][offset]  = noise_T[src_planes[0][offset]  + src_planes[0][offset]  + backup[0][offset]];
                 } else {
                     dst_planes[0][offset]  = src_planes[0][offset];
                 }
             }
             diff = abs(Gs-Go);
             if (diff < value1) {
                 if (diff > (value2)) {
                     dst_planes[1][offset4]  = noise_T[src_planes[1][offset4]  + src_planes[1][offset4]  + backup[1][offset4]];
                 } else {
                     dst_planes[1][offset4]  = src_planes[1][offset4];
                 }
             }
             diff = abs(Bs-Bo);
             if (diff < value1) {
                 if (diff > (value2)) {
                     dst_planes[2][offset4]  = noise_T[src_planes[2][offset4]  + src_planes[2][offset4]  + backup[2][offset4]];
                 } else {
                     dst_planes[2][offset4]  = src_planes[2][offset4];
                 }
             }

             offset++;
             offset4_f += 0.5;

             /* TEST */ //dst_planes[0][offset] = x;
        }
    }

    /* Backup current frame for next iteration */
    fast_memcpy(backup[0],src_planes[0],wh);
    fast_memcpy(backup[1],src_planes[1],wh4);
    fast_memcpy(backup[2],src_planes[2],wh4);

    /* Clean & Exit */
    return dst_planes[0];
}   


G_MODULE_EXPORT void noisefilter_destruct(void) {
    g_log(DRIP_CB_LD,G_LOG_LEVEL_DEBUG,"Plugin: noisefilter destruct");
    /* Clean & Exit */
    return;
}


G_MODULE_EXPORT void noisefilter_init(gint w, gint h,
                                     gint xo, gint yo, gint xl, gint yl,
                                     gfloat fr,
                                     guint8 *src[3], guint8 *dst[3]) {
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Plugin: noisefilter init");
    /* Copy values */
    width = w; height = h;
    xoff = xo; yoff = yo; xlen = xl; ylen = yl;
    framerate = fr;
    src_planes[0] = src[0]; src_planes[1] = src[1]; src_planes[2] = src[2];
    dst_planes[0] = dst[0]; dst_planes[1] = dst[1]; dst_planes[2] = dst[2];
    wh = w*h;
    wh4 = wh/4;
    backup[0] = (guchar*)malloc(wh);
    backup[1] = (guchar*)malloc(wh4);
    backup[2] = (guchar*)malloc(wh4);
    value2 = value1 * 2;

    for (gint i=0; i<766; i++) {
        noise_T[i]=i/3;
    }

    for (gfloat i=-16;i<240;i++) {
        noise_T1164[(gint)(i+16)] = i*1.164;
    }
    for (gfloat i=-128;i<128;i++) {
        noise_T2018[(gint)(i+128)] = i*2.018;
        noise_T0813[(gint)(i+128)] = i*0.813;
        noise_T1596[(gint)(i+128)] = i*1.596;
        noise_T0391[(gint)(i+128)] = i*0.391;
    }
    init_done = TRUE;
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Noise filter Initialised for %ix%i at level %g",width,height,value1);
    return;
}


/* g_module_check_init is automatically executed upon loading */
G_MODULE_EXPORT const gchar* g_module_check_init(GModule *module) {
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Plugin: noisefilter selftest passed");
    return NULL;
}

} //extern "C"

