// ----------------------------------------------------------------------------------
//
// testfilter, skeleton code for building plugins.
//
// 14-03-2002 - Jarl van Katwijk
//
// ----------------------------------------------------------------------------------
//
// Howto code a new plugin:
//
// 1) Define the name of the new plugin, and replace all occurances of 'testfilter'
//    by '{newname}filter'. (Also the C filename).
//    NOTE: check the exsisting filters there aint one with the same name already.
//
// 2) Define the filter type: Audio of Video. Change UNKNOWNT in the type function
//    to the type of the new filter.
// 3) Define the phase where the filter is applied: PRE, NORMAL or POST. For audio
//    filter this is just a sequence: 1st all PRE phase filters are applied, than
//    NORMAL, than POST. For video filters PRE phase filters get unzoomed frames,
//    NORMAL and POST get zoomed frames.
// 4) Implement the filter code. Initialisation in the init function, cleanup\reset
//    code in the destruct function, and the actual filtering in the apply function.
// 5) Edit ./plug-ins/Makefile.am and add the new plugin. And add the new plugin to 
//    ./encoder/plugin-loader.cpp in the function init_plugins(). (Yeah, crappy yet,
//    this will change very soon). Compile, install, and run...
//
//
// ----------------------------------------------------------------------------------


#include "../config.h"
#include "../src/drip.h"
#include "../encoder/plugin-loader.hh"
#include <gmodule.h>
// Custom includes:
// none

static gboolean init_done = FALSE;                 // Plugin initialised flag
static gint width,height;                          // Input Width\Height sizes
static gint xoff,yoff,xlen,ylen;                   // Offset and lenght values
static gfloat framerate;                           // Framerate (25.00, 29.97, etc)
static gfloat value1,value2,value3,value4,value5;  // Plugin defined values - these are loaded\saved by Drip
static guint8 *src_planes[3];                      // Input data planes (I420 colour space)
static guint8 *dst_planes[3];                      // Output data planes
// Custom vari's
// none

extern "C" {





/* ----------------------------------- */
/* FILTER TYPES:  AUDIO, VIDEO, SPU(?) */
/* FILTER PHASE:  PRE, NORMAL, POST    */
/* ----------------------------------- */


/* Return type (SPU,AUDIO,VIDEO) of filter */
G_MODULE_EXPORT module_type_e testfilter_type(void) {
    return UNKNOWNT;                                                      /* DEFINE! */
}


/* Return phase (PRE,NORMAL,POST) of filter */
G_MODULE_EXPORT module_phase_e testfilter_phase(void) {
    return UNKNOWNP;                                                      /* DEFINE! */
}





/* ---------------------------------- */
/* FILTER FUNCTIONS TO BE IMPLEMENTED */
/* ---------------------------------- */

/* Function which implements a GtkPlug to let user configure this plugin */      /* DEFINE! */
G_MODULE_EXPORT GtkWidget* testfilter_gui(void) {
    static GtkWidget *gui = NULL;
    #ifdef DEBUG
    g_log(DRIP_CB_LD,G_LOG_LEVEL_DEBUG,"Plugin: testfilter gui");
    #endif
    /* Already constructed gui? */
    if (gui==NULL) {
        gui = gtk_label_new("\ttestfilter\nThis is just a test filter,\ndoesn't do anything\n");
        /* Build the gtk user interface for this plugins here. Also implement the callbacks */
    }

    /* Clean & Exit */
    return gui;
}


/* This function is called for every (audio\video) frame.
   Here the actual filtering is done. */
G_MODULE_EXPORT guint8* testfilter_apply(guint8* src, glong size, gulong SCR, gulong PTS) {
    #ifdef DEBUG
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Plugin: testfilter apply (PTS=%lu,SCR=%lu)",PTS,SCR);
    #endif


    /* -------------------------- */                                             /* DEFINE! */
    /* IMPLEMENT FILTER CODE HERE */
    /* -------------------------- */


    /* Clean & Exit */
    return src;             // returning input data, here dst=src, but dst!=src is OK too
}


G_MODULE_EXPORT void testfilter_init(gint w, gint h,
                                     gint xo, gint yo, gint xl, gint yl,
                                     gfloat fr,
                                     guint8 *src[3], guint8 *dst[3]) {
    #ifdef DEBUG
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Plugin: testfilter init");
    #endif
    /* Copy values */
    width = w; height = h;
    xoff = xo; yoff = yo; xlen = xl; ylen = yl;
    framerate = fr;
    src_planes[0] = src[0]; src_planes[1] = src[1]; src_planes[2] = src[2];
    dst_planes[0] = dst[0]; dst_planes[1] = dst[1]; dst_planes[2] = dst[2];
    /* Filter specific initialisation */
                                                                                 /* DEFINE! */
    /* Set init flag */
    init_done = TRUE;
    /* Clean & Exit */
    return;
}


/* Called after encoding has ended. Variable can be reset etc. */
G_MODULE_EXPORT void testfilter_destruct(void) {                                 /* DEFINE! */
    #ifdef DEBUG
    g_log(DRIP_CB_LD,G_LOG_LEVEL_DEBUG,"Plugin: testfilter destruct");
    #endif
    /* Clean & Exit */
    return;
}






/* ------------------------------------------------------------------------ */
/* FIXED FUNCTIONS AFTER THIS POINT, DONT EDIT, ONLY CHANGE FUNCTION NAMES! */
/* ------------------------------------------------------------------------ */


/* Return values of plugin (value1,value2,... value5) */
G_MODULE_EXPORT void testfilter_values(gfloat* values) {                          /* DEFINE! */
    values[0] = value1;
    values[1] = value2;
    values[2] = value3;
    values[3] = value4;
    values[4] = value5;
    return;
}

/* Set plugins values */
G_MODULE_EXPORT void testfilter_values_function(gfloat v1,gfloat v2,gfloat v3,gfloat v4,gfloat v5) { /* DEFINE */
    value1 = v1;
    value2 = v2;
    value3 = v3;
    value4 = v4;
    value5 = v5;
    return;
}

/* g_module_check_init is automatically executed upon loading */
G_MODULE_EXPORT const gchar* g_module_check_init(GModule *module) {                /* DEFINE! */
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Plugin: testfilter selftest passed");
    return NULL;
}

} //extern "C"


