#include "../config.h"
#include "../src/drip.h"
#include "../encoder/plugin-loader.hh"
#include <gmodule.h>
// Custom includes:
#include <math.h>

static gboolean init_done = FALSE;                 // Plugin initialised flag
static gint width,height;                          // Input Width\Height sizes
static gint xoff,yoff,xlen,ylen;                   // Offset and lenght values
static gfloat framerate;                           // Framerate (25.00, 29.97, etc)
static gfloat value1,value2,value3,value4,value5;  // Plugin defined values
static guint8 *src_planes[3];                      // Input data planes (I420 colour space)
static guint8 *dst_planes[3];                      // Output data planes
// Custom vari's
// none

/* Gui callback */
void on_spinbutton_volumefilter_changed(GtkEditable *editable,gpointer user_data) {
    GtkWidget *entry = (GtkWidget*)user_data;
    value1 = atof(gtk_entry_get_text(GTK_ENTRY(entry)));
    return;
}

extern "C" {


/* Return type (SPU,AUDIO,VIDEO) of filter */
G_MODULE_EXPORT module_type_e volumefilter_type(void) {
    return AUDIO; // AUDIO filter
}


/* Return phase (PRE,NORMAL,POST) of filter */
G_MODULE_EXPORT module_phase_e volumefilter_phase(void) {
    return POST; // Apply volume adjustment after other phase have been applied
}


/* Return values of plugin (value1,value2,... value5) */
/* NOTE: calling function must deallocate the returned gfloat[5] */
G_MODULE_EXPORT void volumefilter_values(gfloat* values) {
    values[0] = value1;
    values[2] = value2;
    values[3] = value3;
    values[4] = value4;
    values[5] = value5;
    return;
}


/* Set plugins values */
G_MODULE_EXPORT void volumefilter_values_function(gfloat v1,gfloat v2,gfloat v3,gfloat v4,gfloat v5) {
    value1 = v1;
    value2 = v2;
    value3 = v3;
    value4 = v4;
    value5 = v5;
    return;
}


/* Function which implements a GtkWidget to let user configure this plugin */
G_MODULE_EXPORT GtkWidget* volumefilter_gui(void) {
    static GtkWidget *gui = NULL;
    GtkObject *spinbutton_audio_adjustment_adj;
    GtkWidget *spinbutton_audio_adjustment;
    GtkWidget *label_audio_adjustment;

    #ifdef DEBUG
    g_log(DRIP_CB_LD,G_LOG_LEVEL_DEBUG,"Plugin: volumefilter gui");
    #endif
    /* Already constructed gui? */
    if (gui==NULL) {
        
        gui = gtk_hbox_new (FALSE, 0);
        gtk_widget_ref (gui);
        gtk_widget_show (gui);

        label_audio_adjustment = gtk_label_new ("Audio Volume Multiplier");
        gtk_widget_ref (label_audio_adjustment);
        gtk_object_set_data_full (GTK_OBJECT(gui), "label_audio_adjustment", label_audio_adjustment, (GtkDestroyNotify) gtk_widget_unref);
        gtk_widget_show (label_audio_adjustment);
        gtk_box_pack_start (GTK_BOX (gui), label_audio_adjustment, FALSE, FALSE, 0);
        gtk_label_set_justify (GTK_LABEL (label_audio_adjustment), GTK_JUSTIFY_LEFT);
    
        spinbutton_audio_adjustment_adj = gtk_adjustment_new(value1,0,19,1,10,10);
        spinbutton_audio_adjustment = gtk_spin_button_new (GTK_ADJUSTMENT (spinbutton_audio_adjustment_adj), 0.1,3);
        gtk_widget_ref (spinbutton_audio_adjustment);
        gtk_object_set_data_full (GTK_OBJECT(gui), "spinbutton_audio_adjustment", spinbutton_audio_adjustment,
                                  (GtkDestroyNotify) gtk_widget_unref);
        gtk_widget_show (spinbutton_audio_adjustment);
        gtk_box_pack_end (GTK_BOX (gui), spinbutton_audio_adjustment, FALSE, FALSE, 0);
        gtk_widget_set_usize (spinbutton_audio_adjustment, 200, -2);
    
        gtk_signal_connect (GTK_OBJECT (spinbutton_audio_adjustment), "changed",
                            GTK_SIGNAL_FUNC (on_spinbutton_volumefilter_changed),
                            spinbutton_audio_adjustment);
    }

    /* Clean & Exit */
    return gui;
}


G_MODULE_EXPORT guint8* volumefilter_apply(guint8* src, glong size, gulong SCR, gulong PTS) {
    guint16 temp;
    #ifdef DEBUG
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Plugin: volumefilter apply %g",value1);
    #endif
    guint16* src_16 = (guint16*)src;
    for (gint i=0;i<size/2;i++) {
        temp = (guint)(src_16[i] * value1);
        if (temp>65534) temp = 65535;
        src_16[i] = temp;
    }
    /* Clean & Exit */
    return src;             // returning input data: dst=src
}


G_MODULE_EXPORT void volumefilter_init(gint w, gint h,
                                     gint xo, gint yo, gint xl, gint yl,
                                     gfloat fr,
                                     guint8 *src[3], guint8 *dst[3]) {
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Plugin: volumefilter init");
    /* Copy values */
    width = w; height = h;
    xoff = xo; yoff = yo; xlen = xl; ylen = yl;
    framerate = fr;
    src_planes[0] = src[0]; src_planes[1] = src[1]; src_planes[2] = src[2];
    dst_planes[0] = dst[0]; dst_planes[1] = dst[1]; dst_planes[2] = dst[2];
    /* Set init flag */
    init_done = TRUE;

    /* Clean & Exit */
    return;
}


G_MODULE_EXPORT void volumefilter_destruct(void) {
    g_log(DRIP_CB_LD,G_LOG_LEVEL_DEBUG,"Plugin: volumefilter destruct");
    /* Clean & Exit */
    return;
}


/* g_module_check_init is automatically executed upon loading */
G_MODULE_EXPORT const gchar* g_module_check_init(GModule *module) {
    /* Set value1 to volume adjustment value */
    value1 = 1.0;
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Plugin: volumefilter selftest passed");
    return NULL;
}

} //extern "C"
