/*
 * Drip - Jarl van Katwijk
 * 
 * Drip is GPL licenced
 */

#include "../config.h"
#include <gnome.h>
#include <glib.h>
#include "interface.h"
#include "support.h"
#include "drip.h"
#include <dirent.h>
#include <sys/stat.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/types.h>
#include <ctype.h>
#include <fcntl.h>
#include <errno.h>
#include "callbacks.h"
#include "process.h"
#include <sys/time.h>
#include "../encoder/external.hh"
#include "../encoder/plugin-loader.hh"
//#include "../encoder/plugin-loader.hh"
#include "external.h"

//#define DEBUG

/* local proto's */
void avifile_scan(void);

/* globals */
gboolean booted = FALSE;
gboolean logging_locked = FALSE;
GdkFont *logfont;
GdkColor col_debug,col_warning,col_error,col_critical,col_info,col_message;
struct timeval encoding_tv;
struct timezone encoding_tz={0,0};
gboolean encoding_time = FALSE;
gint log_file;
drip_config Config;
GString *status_action;
guint status_perc;
gint linenr = 0;
guint log_handler_drip_id,log_handler_drip_cb_id;
pthread_t serverorbthread;
gint serverorbthread_id;
gboolean server_orb = FALSE;
GtkWidget *mainwindow;
GtkWidget *framewindow = NULL;
GtkWidget *druidtest = NULL;
// gpointer currentframebuffer;
gint xd = 0;
gboolean making_iso;
gboolean encoding;
gboolean encoding_paused;
gboolean stopped;
pthread_t encoderthread;
gint encoderthread_id;
GString *previous_location;
GMutex *LOCKED;
gboolean display_unlock = FALSE;


/* GUI ENTRIES FOR ENCODER IMPLEMENTATIONS */

void gui_ready() {
    /* Done encoding, reset and notice user */
    GString *command = g_string_new("");

    /* Rename generated files */
    rename_divx_files();

    /* Notice user */
    g_log(DRIP_LD,G_LOG_LEVEL_MESSAGE,"Ready, encoding done ----------------------------------------------");

    /* Dealloc progress frame */
    drip_lock("GUI ready1");
    gtk_widget_destroy(gpixmap);gpixmap = NULL;
    drip_unlock("GUI ready1");

    /* Make ISO images out of generated DIVX's */
    if (Config.create_iso == TRUE) {
        pthread_t isothread;
        gint isothread_id;
        making_iso = TRUE;
        isothread_id = pthread_create(&isothread,NULL,make_iso_images,NULL);
        while (making_iso==TRUE) usleep(100000);
    }

    /* Set GUI statusbar */
    drip_lock("GUI ready2");
    gnome_appbar_set_status(GNOME_APPBAR(appbar1),"Ready.");
    gnome_appbar_set_progress(GNOME_APPBAR(appbar1),0);
    drip_unlock("GUI ready2");

    /* Mark converting status */
    converting = FALSE;

    /* Clean & exit */
    g_string_free(command,TRUE);
    return;
}

void gui_progress(glong frame,glong total) { //,glong currentvob,glong totalvobs) {
    /* Display progress to user */
    static glong total_previous = 1000;
    GString *frames = g_string_new("");
    gfloat base = 1;//totalvobs;
    gfloat perc = 1;//totalvobs;
    gfloat part = 0;
    GString *eta;
    struct timeval current_tv;
    struct timezone current_tz={0,0};
    gulong seconds_left = 59;
    gulong minutes_left = 59;
    gulong hours_left = 99;

    //temp stuff:
    glong currentvob = 0;
    //----------

    if (total_previous>total) {
        /* get start time */
        gettimeofday(&encoding_tv,&encoding_tz);
        encoding_time = TRUE;
    }
    total_previous = total;

    if (total>0) {
        /* Calc perc */
        if (perc>0) {
            perc = 1/perc*(currentvob+1);
            if (total>0) {
                part = perc*((gfloat)frame/(gfloat)total);
            }
        }
        if (base>0) {
            base = 1/base*currentvob;
        }

        /* Check if progress makes any sense, correct if not */
        if (perc>1)    { perc = 1;}
        if (part>perc) { part = perc;}
        if (part<base) { part = base;}
        if (part==0)   { part = 0.0000001;};

        /* Build ETA time string */
        if (encoding_time == FALSE) {
            /* get start time */
            gettimeofday(&encoding_tv,&encoding_tz);
            encoding_time = TRUE;
        }
        gettimeofday(&current_tv,&current_tz);
        seconds_left = (gulong)((current_tv.tv_sec - encoding_tv.tv_sec) * (1-part) / part);

        hours_left = (guint)(seconds_left/3600);
        seconds_left -= 3600*hours_left;

        minutes_left = (guint)(seconds_left/60);
        seconds_left -= 60*minutes_left;

        eta = g_string_new("");
        if (hours_left>0) {
            g_string_sprintf(eta,"%li hours %li minutes",hours_left,minutes_left);
        } else {
            g_string_sprintf(eta,"%li minutes %li seconds",minutes_left,seconds_left);
        }

        /* Show */
        g_string_sprintf(frames,"   %.2f%%    Frame %li/%li (%.2ffps)    Remaining %s",part*100,frame,total,(gfloat)(frame/(gfloat)(current_tv.tv_sec - encoding_tv.tv_sec)),eta->str);
        drip_lock("GUI progress1");
        gnome_appbar_set_status(GNOME_APPBAR(appbar1),frames->str);
        gnome_appbar_set_progress(GNOME_APPBAR(appbar1),part);
        drip_unlock("GUI progress1");
        g_string_free(eta,TRUE);
    } else {
        g_string_sprintf(frames,"Encoding frame %li",frame);
        drip_lock("GUI progress2");
        gnome_appbar_set_status(GNOME_APPBAR(appbar1),frames->str);
        gnome_appbar_set_progress(GNOME_APPBAR(appbar1),part);
        drip_unlock("GUI progress2");
    }

    /* Clean & exit */
    g_string_free(frames,TRUE);
    return;
}

/* Show progress frame that's been generate by the backend every N frames */
void gui_frame(gpointer buffer,glong frame,glong x,glong y) {
    gint yd;
    GString *frame_filename = g_string_new("");

    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"New progress frame %li is available (%li,%li)",frame,x,y);
    drip_lock("GUI frame1");

    /* Clean previous frame */
    if (gpixmap!=NULL) {
        gtk_widget_destroy(gpixmap);
    }

    /* Build new pixmap widget */
    xd = hbox_main->allocation.width;                  // ADAPTIVE SIZE

    yd = (gint)(y*((gfloat)xd/x));
    gpixmap = (GtkWidget*)gnome_pixmap_new_from_rgb_d_at_size((guchar*)buffer,(guchar*)NULL,x,y,xd*5/6,yd*5/6); // TODO: ASCPECT
    gtk_box_pack_start (GTK_BOX (hbox_main), gpixmap, FALSE, FALSE, 0);
    gtk_widget_show(gpixmap);

    /* Clean Exit */
    g_string_free(frame_filename,TRUE);
    drip_unlock("GUI frame1");
    return;
}

void gui_errorsignal(glong signal) {
    g_log(DRIP_LD,G_LOG_LEVEL_CRITICAL,"BACKEND: Signal %li has occured! Please send a detailed bug report to the Drip Project",signal);
    return;
}

/* Searches the /etc/fstab for either a cdrom or dvd entry 
 * If it finds one it returns the device for it. If not, 
 * default to /dev/cdrom ( a safe bet )
 */
gchar* detect_dvdrom() {
    GString * device;
    gint result;
    device = g_string_new("");
	
    /* Look for a device in mtab  */
    device = drip_system("grep dvd /etc/fstab | awk '{print $1}'",&result,FALSE,TRUE);
    if (device->len==0 || strstr(device->str,"{empty output}")!=NULL) {
        /* Couldn't find anything. Safe default */
        g_string_sprintf(device,"%s","/dev/cdrom");
    }
    return device->str;
}

void log_handler(const gchar *log_domain,GLogLevelFlags log_level,const gchar *message,gpointer user_data) {
    /* Handles logging. See glib logging system for more info */
    time_t Time = time(NULL);
    GString *date;
    GString *domain;
    GString *level;
    GString *logstring;
    GString *displaystring;
    gboolean quit;
    gint bytes = 0;
    gchar *last_char;
    GdkColor *colour = NULL;

    /* Check log domain */
    if (strcmp(log_domain,DRIP_LD)!=0 && strcmp(log_domain,DRIP_CB_LD)!=0) {
        g_error("Got a log for domain %s, which does not belong to the drip log domain.",log_domain);
    }

    /* Lock if not logging from a CallBack function */
    if (strcmp(log_domain,DRIP_LD)==0) {
        drip_lock("log handler1");
    }

    /* Format domain */
    domain = g_string_new(log_domain);
    while (domain->len < 8) {
        domain = g_string_append(domain," ");
    }

    /* Handle log_level, strip flags */
    log_level = (GLogLevelFlags)(log_level & 252);
    switch (log_level) {
        case G_LOG_LEVEL_ERROR:      level = g_string_new("Error");
                                     colour = &col_error;
                                     quit = TRUE;
                                     break;
        case G_LOG_LEVEL_CRITICAL:   level = g_string_new("Critical");
                                     colour = &col_critical;
                                     quit = TRUE;
                                     break;
        case G_LOG_LEVEL_WARNING:    level = g_string_new("Warning");
                                     colour = &col_warning;
                                     quit = FALSE;
                                     break;
        case G_LOG_LEVEL_MESSAGE:    level = g_string_new("Message");
                                     colour = &col_message;
                                     quit = FALSE;
                                     break;
        case G_LOG_LEVEL_INFO:       level = g_string_new("Info");
                                     colour = &col_info;
                                     quit = FALSE;
                                     break;
        case G_LOG_LEVEL_DEBUG:      level = g_string_new("Debug");
                                     colour = &col_debug;
                                     quit = FALSE;
                                     break;
        default:                     level = g_string_new("Unknown");
                                     quit = FALSE;
                                     break;
    }

    /* Text formatting */
    while (level->len < 8) {
        level =  g_string_append(level," ");
    }

    /* Format date string */
    date = g_string_new(ctime(&Time));
    date = g_string_truncate(date,(date->len)-1);

    /* Store log entry */
    logstring = g_string_new("");
    g_string_sprintf(logstring,"%s | %i\t| %s | %s\n",date->str,linenr,level->str,message);
    displaystring = g_string_new("");
    g_string_sprintf(displaystring,"%s | %s\n",date->str,message);
    last_char = (gchar*)((glong)logstring->str + logstring->len - 2); /* we got 2x \n? */
    if (*last_char == 10) {
        logstring = g_string_truncate(logstring,logstring->len -1);
    }
    bytes = write(log_file,logstring->str,logstring->len); /* write log entry to log file */

    /* Write criticals to tty and new dialog window */
#ifndef DEBUG
    if (log_level == G_LOG_LEVEL_CRITICAL) {
        drip_info(logstring->str);
#endif
        g_print(logstring->str);
#ifndef DEBUG
    }
#endif

    /* Log to GUI */
    if (textscreen!=NULL) {
        if (log_level!=G_LOG_LEVEL_DEBUG || Config.debug_log==TRUE) {
            gtk_text_insert (GTK_TEXT(textscreen), logfont, colour, NULL, displaystring->str, -1);
        }
    }
    linenr++;

    /* Should we exit Drip? */
    if (quit) {
        /* Lock if not logging from a CallBack function */
        if (strcmp(log_domain,DRIP_LD)==0) {
            drip_unlock("log handler1");
        }
        drip_info(logstring->str);
        g_print("sleeping...");
        sleep(10);
        g_print(" done\n");
        terminate(TRUE,FALSE);
    }

    /* Free strings */
    g_string_free(date,TRUE);
    g_string_free(domain,TRUE);
    g_string_free(level,TRUE);
    g_string_free(logstring,TRUE);
    g_string_free(displaystring,TRUE);
    logging_locked = FALSE;
#ifdef DEBUG
    fflush(NULL);
#endif
    /* Lock if not logging from a CallBack function */
    if (strcmp(log_domain,DRIP_LD)==0) {
        drip_unlock("log handler1");
    }

    return;
}


/* Traceback if Config.DVDdevice is a symbolic link */
gboolean DVDsymlink(void) {
    GString *command = g_string_new("");
    guchar *symlink = NULL;
    gint n;
    do {
        symlink = (guchar*)malloc(256);
        n = readlink(Config.DVDdevice->str,symlink,256);
        if (n!=-1 && n<255) { /* DVDdevice is a symlink, use real device instead */
            symlink[n] = 0; /* NULL terminate */
            g_string_sprintf(command,"Using DVD device %s because %s is a symlink",symlink,Config.DVDdevice->str);
            g_log(DRIP_LD,G_LOG_LEVEL_MESSAGE,command->str);
            g_string_sprintf(Config.DVDdevice,"%s",symlink);
            free(symlink);
        } else if (errno!=22) { /* Err=22 -> end looping, real device found */
            g_log(DRIP_LD,G_LOG_LEVEL_ERROR,"Error while parsing DVD device |%s| : %s (%i)",Config.DVDdevice->str,strerror(errno),errno);
        }
    } while (n>-1);

    g_string_free(command,TRUE);
    return TRUE;
}


/* Initialise Drip */
void drip_init() {
    GString *configfile;
    GString *HOME;
    GIOChannel *configfile_channel = NULL;
    glong configfile_size;
    guint bytesloaded;
    gint loaderror;
    gchar *buf;
    gint fd;
    gint i,n;
    glong bytes_to_read;
    struct stat file_stats;
    gboolean true_udf,dummy_bool,config_ok;
    GString *command = g_string_new("");
    GString *logfile = g_string_new("");
    GString *logfile_old = g_string_new("");
    GtkWidget *welcome;

    LOCKED = g_mutex_new();
    previous_location = g_string_new("");
    HOME = g_string_new(getenv("HOME"));
    /* Scan avifile for loaded codecs */
    avifile_scan();
    /* Initialize all drip defaults */
    g_string_sprintf(command,"%s/.drip",HOME->str);
    mkdir (command->str, S_IRUSR | S_IWUSR | S_IXUSR);
    g_string_sprintf(command,"%s/.drip/logs_old",HOME->str);
    mkdir (command->str, S_IRUSR | S_IWUSR | S_IXUSR);
    g_string_sprintf(command,"%s/.drip/plugins",HOME->str);
    mkdir (command->str, S_IRUSR | S_IWUSR | S_IXUSR);

    status_action = g_string_new("");
    configfile = g_string_new("");
    g_string_sprintf(configfile,"%s/.drip/drip.cfg",getenv("HOME"));
    for (n=0;n<256;n++) {
        for (i=0;i<8;i++) { Config.audio_channel_names[n][i] = g_string_new("");}
        for (i=0;i<32;i++) { Config.subpicture_names[n][i] = g_string_new("");}
    }

    /* Logging font and colours */
    logfont = gdk_font_load("-Schumacher-Clean-Medium-R-Normal--14-140-75-75-C-70-ISO646.1991-IRV");
    gdk_color_parse("midnight blue",&col_debug);
    gdk_color_parse("DarkRed",&col_warning);
    gdk_color_parse("tomato3",&col_error);
    gdk_color_parse("Red",&col_critical);
    gdk_color_parse("DarkSlateGrey",&col_info);
    gdk_color_parse("black",&col_message);

    /* Default dvd related settings */
    Config.DVDdevice = g_string_new("");
    Config.DVDsource = g_string_new("");
    Config.DIVXfileUser = g_string_new("");
    Config.DIVXfileFormatted = g_string_new("");
    Config.track_dir = g_string_new("");
    Config.CACHElocation = g_string_new("");
    for (n=0;n<256;n++) {
        Config.audio_channels_in_dvd[n] = 0;
        Config.subpictures_in_dvd[n] = -1;
    }
    Config.subpicture_stream = -1;
    Config.clutfile = g_string_new("");

    /* DVD volume settings */
    Config.DVDvolumeid = g_string_new("");
    Config.DVDpublisher = g_string_new("");
    Config.DVDpreparer = g_string_new("");
    Config.DVDcreationdate = g_string_new("");
    Config.DVDvolumesize_MB = 0;

    /* Drip logging domain */
    g_string_sprintf(logfile,"%s/.drip/drip.log",getenv("HOME"));
    if (stat(logfile->str,&file_stats)!=-1) {
        g_string_sprintf(logfile_old,"%s/.drip/logs_old/drip.%li.log",getenv("HOME"),file_stats.st_ctime);
        rename(logfile->str,logfile_old->str);
    } else {
        g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Could not stat log file, removed it");
        unlink(logfile->str); /* delete old log file, TODO: log file rotation */
    }
    log_file = open(logfile->str,O_WRONLY|O_CREAT,S_IREAD|S_IWRITE); /* create log file */
    log_handler_drip_id = g_log_set_handler (DRIP_LD, (GLogLevelFlags)(G_LOG_LEVEL_WARNING |G_LOG_LEVEL_MESSAGE | G_LOG_LEVEL_INFO | G_LOG_LEVEL_DEBUG | G_LOG_LEVEL_CRITICAL | G_LOG_LEVEL_ERROR | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION),log_handler, NULL);
    log_handler_drip_cb_id = g_log_set_handler (DRIP_CB_LD, (GLogLevelFlags)(G_LOG_LEVEL_WARNING |G_LOG_LEVEL_MESSAGE | G_LOG_LEVEL_INFO | G_LOG_LEVEL_DEBUG | G_LOG_LEVEL_CRITICAL | G_LOG_LEVEL_ERROR | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION),log_handler, NULL);

    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Activated log handler and created log file %s",logfile->str);
    g_print("Activated log handler and created log file %s\n-------------------------------------------------------------------\n",logfile->str);

    /* Build main window */
    mainwindow = create_mainwindow();
    gtk_widget_show (mainwindow);

    /* Any saved configuration? */
    config_ok = FALSE;
    fd = open(configfile->str,O_RDONLY);
    if (fd>-1) {
        configfile_channel = g_io_channel_unix_new(fd);
        stat(configfile->str, &file_stats);
        configfile_size = file_stats.st_size;
        if (configfile_size>0) {
            buf = (gchar*)malloc(configfile_size);
            memset(buf,0,configfile_size);
            /* Check config version */
            g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Checking config version");
            loaderror = g_io_channel_read(configfile_channel,(gchar*)&bytes_to_read,sizeof(glong),&bytesloaded);
            loaderror = g_io_channel_read(configfile_channel,buf,bytes_to_read,&bytesloaded);
            
            if (strstr(CONFIG_VERSION,buf)==NULL) {
                g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Saved configuration (%s) is of incompatible version (%s)",buf,CONFIG_VERSION);
                free(buf);
                g_io_channel_close(configfile_channel);
                close(fd);
            } else {
                /* Read configuration from disk */
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&bytes_to_read,sizeof(glong),&bytesloaded);
                loaderror = g_io_channel_read(configfile_channel,buf,bytes_to_read,&bytesloaded);
                    g_string_sprintf(Config.DVDdevice,"%s",(gchar*)buf);
                    memset(buf,0,configfile_size);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&bytes_to_read,sizeof(glong),&bytesloaded);
                loaderror = g_io_channel_read(configfile_channel,buf,bytes_to_read,&bytesloaded);
                    g_string_sprintf(Config.DIVXfileUser,"%s",(gchar*)buf);
                    memset(buf,0,configfile_size);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&Config.width,sizeof(glong),&bytesloaded);
                    memset(buf,0,configfile_size);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&Config.height,sizeof(glong),&bytesloaded);
                    memset(buf,0,configfile_size);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&Config.video_bitrate,sizeof(guint),&bytesloaded);
                    memset(buf,0,configfile_size);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&Config.video_stream,sizeof(guint),&bytesloaded);
                    memset(buf,0,configfile_size);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&Config.audio_bitrate,sizeof(gint),&bytesloaded);
                    memset(buf,0,configfile_size);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&Config.audio_stream,sizeof(guint),&bytesloaded);
                    memset(buf,0,configfile_size);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&Config.encoding_video_codec,sizeof(gint),&bytesloaded);
                if (Config.codecs_amount<=Config.encoding_video_codec) {
                    Config.encoding_video_codec = 0;
                    Config.encoding_audio_codec = 0;
                    g_log(DRIP_LD,G_LOG_LEVEL_INFO,"The plugins of avifile have changed, resetted codecs");
                }
                    memset(buf,0,configfile_size);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&Config.cliptop,sizeof(gint),&bytesloaded);
                    memset(buf,0,configfile_size);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&Config.clipbottom,sizeof(gint),&bytesloaded);
                    memset(buf,0,configfile_size);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&Config.blurradius,sizeof(gint),&bytesloaded);
                    memset(buf,0,configfile_size);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&dummy_bool,sizeof(gboolean),&bytesloaded);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&Config.eject_dvdrom,sizeof(gboolean),&bytesloaded);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&Config.use_dvddb,sizeof(gboolean),&bytesloaded);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&Config.nice,sizeof(gint),&bytesloaded);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&Config.conversion_mode,sizeof(cmode_t),&bytesloaded);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&Config.maxsize,sizeof(glong),&bytesloaded);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&true_udf,sizeof(gboolean),&bytesloaded);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&Config.encode,sizeof(gboolean),&bytesloaded);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&bytes_to_read,sizeof(glong),&bytesloaded);
                loaderror = g_io_channel_read(configfile_channel,buf,bytes_to_read,&bytesloaded);
                    g_string_sprintf(Config.DVDsource,"%s",(gchar*)buf);
                    memset(buf,0,configfile_size);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&Config.useDVDdevice,sizeof(gboolean),&bytesloaded);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&Config.cpus,sizeof(gint),&bytesloaded);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&Config.done_warning,sizeof(gboolean),&bytesloaded);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&bytes_to_read,sizeof(glong),&bytesloaded);
                loaderror = g_io_channel_read(configfile_channel,buf,bytes_to_read,&bytesloaded);
                    g_string_sprintf(Config.CACHElocation,"%s",(gchar*)buf);
                    memset(buf,0,configfile_size);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&Config.cache_delete_on_exit,sizeof(gboolean),&bytesloaded);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&Config.create_iso,sizeof(gboolean),&bytesloaded);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&Config.hyper_zoom,sizeof(gboolean),&bytesloaded);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&Config.autoclip,sizeof(gboolean),&bytesloaded);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&Config.clipright,sizeof(gint),&bytesloaded);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&Config.clipleft,sizeof(gint),&bytesloaded);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&dummy_bool,sizeof(gboolean),&bytesloaded);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&Config.deinterlace,sizeof(gboolean),&bytesloaded);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&Config.pulldown,sizeof(gboolean),&bytesloaded);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&Config.encoding_audio_codec,sizeof(gint),&bytesloaded);
                if (Config.codecs_amount<=Config.encoding_audio_codec) {
                    Config.encoding_audio_codec = 0;
                    Config.encoding_video_codec = 0;
                    g_log(DRIP_LD,G_LOG_LEVEL_INFO,"The plugins of avifile have changed, resetted codecs");
                }

                loaderror = g_io_channel_read(configfile_channel,(gchar*)&Config.mainwindow_X,sizeof(gint),&bytesloaded);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&Config.mainwindow_Y,sizeof(gint),&bytesloaded);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&Config.debug_log,sizeof(gboolean),&bytesloaded);
                loaderror = g_io_channel_read(configfile_channel,(gchar*)&Config.audio_norm,sizeof(gboolean),&bytesloaded);

                free(buf);
                g_io_channel_close(configfile_channel);
                close(fd);
                config_ok = TRUE;
                g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Loaded saved configuration");
            }
        }
    } 

    if (config_ok == FALSE) {
        /* Default divx related settings */
        g_log(DRIP_LD,G_LOG_LEVEL_INFO,"Setting up default configuration...");
        g_string_sprintf(Config.DVDdevice,"%s","/dev/cdrom");
        /* Traceback if Config.DVDdevice is a symbolic link */
        DVDsymlink();

        g_string_sprintf(Config.DVDsource,"./cache/");
        g_string_sprintf(Config.DIVXfileUser,"drip.avi");
        g_string_sprintf(Config.CACHElocation,"./cache/");
        Config.width = 512;
        Config.height = -1;
        Config.hyper_zoom = TRUE;
        Config.video_bitrate = 800;
        Config.video_stream = 0;
        Config.audio_bitrate = 64;
        Config.audio_stream = 0;
        Config.audio_norm = TRUE;
        /* Video codec, scan for 1st valid codec */
        for (i=0;i<Config.codecs_amount;i++) {
            if (Config.codecs[i].cav == VIDEO_CODEC) {
                Config.encoding_video_codec = i;
                i = Config.codecs_amount;
            }
        } 
        /* Audio codec, scan for 1st valid codec */
        for (i=0;i<Config.codecs_amount;i++) {
            if (Config.codecs[i].cav == AUDIO_CODEC) {
                Config.encoding_audio_codec = i;
                i = Config.codecs_amount;
            }
        } 
        Config.cliptop = 0;
        Config.clipbottom = 0;
        Config.autoclip = TRUE;
        Config.deinterlace = FALSE;
        Config.pulldown = FALSE;
        Config.clipright = 0;
        Config.clipleft = 0;
        Config.blurradius = 0;
        /* Default various other settings */
        Config.cache_delete_on_exit = FALSE;
        Config.eject_dvdrom = TRUE;
        Config.nice = 0;
        Config.conversion_mode = UNKNOWN;
        Config.use_dvddb = FALSE;
        Config.maxsize = 696;
        Config.encode = TRUE;
        Config.useDVDdevice = TRUE;
        Config.cpus = 1;
        Config.done_warning = TRUE;
        Config.create_iso = FALSE;
        Config.mainwindow_X = 500;
        Config.mainwindow_Y = 400;
#ifdef DEBUG
        Config.debug_log = TRUE;
#else
        Config.debug_log = FALSE;
#endif
        /* No save config around? -> Welcome user */
        welcome = create_messagebox_betanotice();
        gtk_widget_show(welcome);
    }

    /* Load plugins */
    if (init_plugins() == FALSE) {
        g_log(DRIP_LD,G_LOG_LEVEL_ERROR,"Plugins not loaded");
    }

    /* Clean & Exit */
    if (Config.CACHElocation->len==0) {
        /* check if cache location is empty */
        g_string_sprintf(Config.CACHElocation,"./");
    }
    g_string_free(HOME,TRUE);
    g_string_free(logfile,TRUE);
    g_string_free(logfile_old,TRUE);
    g_string_free(configfile,TRUE);
    g_string_free(command,TRUE);
    return;
}

/* Lock */
void drip_lock(const gchar *location) {
    /* Already locked? */
    if (g_mutex_trylock(LOCKED) == FALSE) {
        display_unlock = TRUE;
        printf("Already locked from %s, waiting to become unlocked for %s\n",previous_location->str,location);
        while (g_mutex_trylock(LOCKED)==FALSE) {
            usleep(100);
        }
        printf("Engine got unlocked from %s and locked for %s\n",previous_location->str,location);
    }
    /* Lock gdk */
    g_string_sprintf(previous_location,"%s",location);
    gdk_threads_enter();
#ifdef DEBUG
    printf("Locked, from %s\n",location);
#endif
    return;
}

/* Unlock */
void drip_unlock(const gchar *location) {
    /* Unlocking from correct location? */
    if (strcmp(previous_location->str,location)!=0) {
        /* Display termination reason, DONT use g_log(DRIP_LD,..) for this! */
        printf("Unlocking from incorrect location %s, %s was exspected, this should not happened!\n",location,previous_location->str);
        terminate(TRUE,FALSE);
    }
    /* Unlock */
    g_mutex_unlock(LOCKED);
    gdk_threads_leave();
    if (display_unlock == TRUE) {
        display_unlock = FALSE;
#ifdef DEBUG
        printf("Unlocked from %s\n",location);
#endif
    }
#ifdef DEBUG
    else {
        printf("Unlocked from %s\n",location);
    }
#endif
    return;
}

/* Terminates drip, should do clean shutdown  */
void terminate(gboolean hard,gboolean lock) {
    GString *str = g_string_new("");
    GString *command = g_string_new("");

    if (converting && hard==FALSE) {
        GtkWidget* buzy;
        buzy = create_messagebox_buzy();
        gtk_widget_show(buzy);
    } else {
        close_plugins();
        save_config(FALSE);
        remove_cache(FALSE);
        gdk_threads_leave(); // Overrule locks
        g_log_remove_handler(DRIP_LD,log_handler_drip_id);
        close(log_file);
        encoder_quit(TRUE);
        _exit(0);
    }

    /* Clean & Exit */
    g_string_free(str,TRUE);
    g_string_free(command,TRUE);
    return;
}

void avifile_scan() {
    codecS** codec_list;
    guint i;
    /* Fetch available output codecs */
    Config.codecs = (codecS*)malloc(sizeof(codecS)*256);
    codec_list = encoder_codecs();
    Config.codecs_amount = 0;
    for (i=0;i<codec_list[0]->length;i++) {
        Config.codecs[Config.codecs_amount].ctype = g_string_new("");
        Config.codecs[Config.codecs_amount].cmodule = g_string_new("");
        Config.codecs[Config.codecs_amount].cdescription = g_string_new("");
        if (codec_list[i]->ctype->len>0 && isupper(codec_list[i]->ctype->str[0])) {
            /* Fill local config */
            Config.codecs[Config.codecs_amount].ctype = g_string_assign(Config.codecs[Config.codecs_amount].ctype,codec_list[i]->ctype->str);
            Config.codecs[Config.codecs_amount].ccodec = codec_list[i]->ccodec;
            Config.codecs[Config.codecs_amount].cmodule = g_string_assign(Config.codecs[Config.codecs_amount].cmodule,codec_list[i]->cmodule->str);
            /* -- Setup description */
            /* Video codecs */
            if (strcmp(Config.codecs[Config.codecs_amount].ctype->str,"DIV3")==0) {
                Config.codecs[Config.codecs_amount].cdescription = g_string_assign(Config.codecs[Config.codecs_amount].cdescription,"MSMPEG4 low motion codec");
            }
            if (strcmp(Config.codecs[Config.codecs_amount].ctype->str,"DIV4")==0) {
                Config.codecs[Config.codecs_amount].cdescription = g_string_assign(Config.codecs[Config.codecs_amount].cdescription,"MSMPEG4 high motion codec");
            }
            if (strcmp(Config.codecs[Config.codecs_amount].ctype->str,"DIVX")==0) {
                Config.codecs[Config.codecs_amount].cdescription = g_string_assign(Config.codecs[Config.codecs_amount].cdescription,"OpenDIVX codec");
            }
            /* Audio codecs */
            if (strcmp(Config.codecs[Config.codecs_amount].ctype->str,"P")==0) {
                Config.codecs[Config.codecs_amount].cdescription = g_string_assign(Config.codecs[Config.codecs_amount].cdescription,"?Vorbis? codec");
            }
            if (strcmp(Config.codecs[Config.codecs_amount].ctype->str,"U")==0) {
                Config.codecs[Config.codecs_amount].cdescription = g_string_assign(Config.codecs[Config.codecs_amount].cdescription,"MP3 codec");
            }
            switch (codec_list[i]->cav) {
                case 0: Config.codecs[Config.codecs_amount].cav = VIDEO_CODEC;
                        break;
                case 1: Config.codecs[Config.codecs_amount].cav = AUDIO_CODEC;
                        break;
                default:Config.codecs[Config.codecs_amount].cav = UNKNOWN_CODEC;
                        break;
            }

            /* Unknown codecs */
            if (Config.codecs[Config.codecs_amount].cdescription->len==0) {
                // mouarf..
            } else {
                /* Known codecs, increase amount */
                Config.codecs_amount++;
            }
        }
    }

    /* Check Config.encoding_video_codec & Config.encoding_audio_codec are within range */
    if (Config.encoding_video_codec>=Config.codecs_amount) {
        Config.encoding_video_codec = 0;
    }
    if (Config.encoding_audio_codec>=Config.codecs_amount) {
        Config.encoding_audio_codec = 0;
    }

    return;
}


int main (int argc, char *argv[]) {

    /* Init's */
    g_thread_init(NULL);
    gtk_init(&argc,&argv);
    gnome_init ("gnomedrip", VERSION, argc, argv);
#ifdef ENABLE_NLS
    bindtextdomain (PACKAGE, PACKAGE_LOCALE_DIR);
    textdomain (PACKAGE);
#endif
    drip_init();
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Drip %s, pid = %i",VERSION,getpid());

    /* We can log to the GUI now, show some info */
    g_log(DRIP_LD,G_LOG_LEVEL_MESSAGE,"Welcome to Drip %s",VERSION);

    /* Start glib */
    g_log(DRIP_LD,G_LOG_LEVEL_DEBUG,"Frontend running inner loop");
    gdk_threads_enter();
    gtk_main();
    gdk_threads_leave();
    return 0;
}

